from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Type

import pcbnew

from .kle_serial import ViaKeyboard


@dataclass
class Footprint:
    libname: str
    name: str

    @classmethod
    def from_str(cls: Type[Footprint], arg: str) -> Footprint:
        parts = arg.rsplit(":", 1)
        if len(parts) != 2:
            msg = f"Unexpected footprint value: `{arg}`"
            raise RuntimeError(msg)
        return Footprint(libname="".join(parts[0]), name=parts[1])

    def load(self) -> pcbnew.FOOTPRINT:
        return pcbnew.FootprintLoad(self.libname, self.name)


class BoardBuilder:
    def __init__(
        self,
        *,
        switch_footprint: str,
        diode_footprint: str,
    ) -> None:
        self.switch_footprint = Footprint.from_str(switch_footprint)
        self.diode_footprint = Footprint.from_str(diode_footprint)

        self.board = pcbnew.CreateEmptyBoard()
        self.nets: dict[str, pcbnew.NETINFO_ITEM] = {}
        self.net_info = self.board.GetNetInfo()
        self.net_count = self.board.GetNetCount()

    def add_footprint(self, footprint: pcbnew.FOOTPRINT) -> pcbnew.FOOTPRINT:
        logging.info(f"Add {footprint.GetReference()} footprint")
        self.board.Add(footprint)
        return footprint

    def add_switch_footprint(self, ref_count: int) -> pcbnew.FOOTPRINT:
        f = self.switch_footprint.load()
        f.SetReference(f"SW{ref_count}")
        return self.add_footprint(f)

    def add_diode_footprint(self, ref_count: int) -> pcbnew.FOOTPRINT:
        f = self.diode_footprint.load()
        f.SetReference(f"D{ref_count}")
        return self.add_footprint(f)

    def add_net(self, netname: str) -> pcbnew.NETINFO_ITEM:
        """Add new net with netname if it does not exist already"""
        if netname in self.nets:
            net = self.nets[netname]
        else:
            net = pcbnew.NETINFO_ITEM(self.board, netname, self.net_count)
            self.net_info.AppendNet(net)
            logging.info(f"Add {netname} net")
            self.board.Add(net)
            self.net_count += 1
            self.nets[netname] = net
        return net

    def create_board(self, keyboard: ViaKeyboard) -> pcbnew.BOARD:
        current_ref = 1

        for key in keyboard.keys:
            row, column = key.labels[0].split(",")

            switch = self.add_switch_footprint(current_ref)
            diode = self.add_diode_footprint(current_ref)

            net = self.add_net(f"COL{column}")
            switch.FindPadByNumber("1").SetNet(net)

            net = self.add_net(f"ROW{row}")
            diode.FindPadByNumber("1").SetNet(net)

            net = self.add_net(f"Net-(D{current_ref})-Pad2")
            switch.FindPadByNumber("2").SetNet(net)
            diode.FindPadByNumber("2").SetNet(net)

            current_ref += 1

        return self.board



from __future__ import annotations

import itertools
import logging
import os
import re
from pathlib import Path
from typing import List, Optional, Tuple, cast

import pcbnew

from .board_modifier import (
    KICAD_VERSION,
    BoardModifier,
    get_closest_pads_on_same_net,
    get_common_nets,
    get_footprint,
    get_optional_footprint,
    get_orientation,
    get_position,
    get_side,
    position_in_rotated_coordinates,
    reset_rotation,
    rotate,
    set_position,
    set_rotation,
    set_side,
)
from .element_position import ElementInfo, ElementPosition, Point, PositionOption
from .kle_serial import Keyboard, get_keyboard


class SwitchIterator:
    def __init__(self, board: pcbnew.BOARD, annotation: str) -> None:
        self.__board = board
        self.__annotation = annotation

    def __iter__(self):
        self.__current_key = 1
        return self

    def __next__(self):
        reference = self.__annotation.format(self.__current_key)
        footprint = self.__board.FindFootprintByReference(reference)
        if footprint:
            result = self.__current_key, footprint
            self.__current_key += 1
            return result
        else:
            raise StopIteration


class KeyPlacer(BoardModifier):
    def __init__(
        self,
        board: pcbnew.BOARD,
        key_distance: Tuple[float, float] = (19.05, 19.05),
    ) -> None:
        super().__init__(board)

        self.__key_distance_x = cast(int, pcbnew.FromMM(key_distance[0]))
        self.__key_distance_y = cast(int, pcbnew.FromMM(key_distance[1]))

        logging.debug(
            f"Set key 1U distance: {self.__key_distance_x}/{self.__key_distance_y}"
        )
        self.__reference_coordinate = pcbnew.wxPointMM(25, 25)

    def route_switch_with_diode(
        self,
        switch: pcbnew.FOOTPRINT,
        diode: pcbnew.FOOTPRINT,
        angle: float,
        template_connection: List[pcbnew.PCB_TRACK] | None = None,
    ) -> None:
        """Performs routing between switch and diode elements.
        It uses two closest (to each other) pads of the same net.

        :param switch: Switch footprint to be routed.
        :param diode: Diode footprint to be routed.
        :param angle: Rotation angle (in degrees) of switch footprint
                      (diode rotation is assumed to be the same)
        :param template_connection: List of template elements (tracks and vias) for
                                    routing switch and diode pads. Normalised to
                                    switch position coordinate. Templates
                                    items must not have netcodes assigned.
                                    If None, use automatic routing algorithm.
        :type switch: FOOTPRINT
        :type diode: FOOTPRINT
        :type angle: float
        :type template_connection: List[pcbnew.PCB_TRACK] | None
        """
        logging.info(f"Routing {switch.GetReference()} with {diode.GetReference()}")

        if template_connection:
            logging.info("Using template replication method")
            if angle != 0:
                logging.info(f"Routing at {angle} degree angle")
            switch_position = get_position(switch)
            rejects = []
            for item in template_connection:
                # item is either PCB_TRACK or PCB_VIA, since via extends track
                # we should be safe with `Cast_to_PCB_TRACK` (not doing any via
                # specific operations here)
                track = pcbnew.Cast_to_PCB_TRACK(item)
                new_track = track.Duplicate()
                if KICAD_VERSION >= (7, 0, 0):
                    new_track.Move(
                        pcbnew.VECTOR2I(switch_position.x, switch_position.y)
                    )
                else:
                    new_track.Move(switch_position)
                if angle != 0:
                    rotate(new_track, switch_position, angle)
                result = self.add_track_to_board(new_track)
                # depending on the order of track elements placement, some may not pass
                # collision check (for example when starting from middle segment,
                # if it ends to close to a pad). To avoid rejecting false positives,
                # collect rejects and give them a second chance. Only the 'middle' segments
                # (i.e. not starting in a pad) should be in this list, others should
                # get placed properly (unless indeed colliding with some footprints,
                # for example optional stabilizer holes), so running placement of rejects
                # a second time should succeed.
                if result is None:
                    rejects.append(new_track)
            for item in rejects:
                self.add_track_to_board(item)
        elif result := get_closest_pads_on_same_net(switch, diode):
            logging.info("Using internal autorouter method")
            switch_pad, diode_pad = result
            self.route(switch_pad, diode_pad)
        else:
            logging.error("Could not find pads with the same net, routing skipped")

    def get_current_relative_element_position(
        self, element1: pcbnew.FOOTPRINT, element2: pcbnew.FOOTPRINT
    ) -> ElementPosition:
        pos1 = get_position(element1)
        pos2 = get_position(element2)
        x = cast(float, pcbnew.ToMM(pos2.x - pos1.x))
        y = cast(float, pcbnew.ToMM(pos2.y - pos1.y))
        return ElementPosition(
            Point(x, y),
            element2.GetOrientationDegrees(),
            get_side(element2),
        )

    def remove_dangling_tracks(self) -> None:
        logging.info("Removing dangling tracks")
        connectivity = self.get_connectivity()

        any_removed = False

        def _is_dangling(track):
            if KICAD_VERSION >= (7, 0, 7):
                return connectivity.TestTrackEndpointDangling(track, False)
            return connectivity.TestTrackEndpointDangling(track)

        for track in self.board.GetTracks():
            if _is_dangling(track):
                logging.info(f"Removing {track.m_Uuid.AsString()}")
                self.board.RemoveNative(track)
                any_removed = True

        if any_removed:
            self.remove_dangling_tracks()

    def save_connection_template(
        self,
        switch: pcbnew.FOOTPRINT,
        diode: pcbnew.FOOTPRINT,
        connections: List[pcbnew.PCB_TRACK],
        destination_path: str,
    ) -> None:
        logging.info(f"Saving template to {destination_path}")
        # can't use `CreateEmptyBoard` when running inside KiCad.
        # We want new board file but without new project file,
        # looks like this is not possible with pcbnew API.
        # So create board with project
        board = pcbnew.NewBoard(destination_path)
        # and delete project file
        os.remove(Path(destination_path).with_suffix(".kicad_pro"))

        switch_copy = pcbnew.Cast_to_FOOTPRINT(switch.Duplicate())
        set_position(switch_copy, pcbnew.wxPoint(0, 0))

        origin = get_position(switch)
        diode_copy = pcbnew.Cast_to_FOOTPRINT(diode.Duplicate())
        set_position(
            diode_copy,
            get_position(diode_copy) - pcbnew.wxPoint(origin.x, origin.y),
        )

        for p in itertools.chain(switch_copy.Pads(), diode_copy.Pads()):
            if p.GetNetCode() != 0:
                logging.info(
                    f"Adding net {p.GetNetname()} with netcode {p.GetNetCode()}"
                )
                # adding nets to new board will get them new autoassigned netcodes
                # (the one set at `NETINFO_ITEM` constructor will be discarded)
                net = pcbnew.NETINFO_ITEM(board, p.GetNetname(), -1)
                board.Add(net)

        nets = board.GetNetsByName()
        # synchronize net codes
        for p in itertools.chain(switch_copy.Pads(), diode_copy.Pads()):
            if p.GetNetCode() != 0:
                before = p.GetNetCode()
                p.SetNet(nets[p.GetNetname()])
                logging.info(
                    f"Updating pad '{p.GetParentAsString()}:{p.GetPadName()}' "
                    f"net {p.GetNetname()} netcode: {before} -> {p.GetNetCode()}"
                )

        board.Add(switch_copy)
        board.Add(diode_copy)

        for item in connections:
            # using `Duplicate` here to not alter net assignments of original tracks
            # (which needs to be empty in order to work as template)
            board.Add(item.Duplicate())
        pcbnew.SaveBoard(destination_path, board, aSkipSettings=True)

    def get_connection_template(
        self, key_format: str, diode_format: str, destination_path: str
    ) -> List[pcbnew.PCB_TRACK]:
        switch = get_footprint(self.board, key_format.format(1))
        diode = get_optional_footprint(self.board, diode_format.format(1))
        if not diode:
            return []

        result = []
        origin = get_position(switch)

        connectivity = self.get_connectivity()
        common_nets = get_common_nets(switch, diode)

        def _append_normalized_connection_items(netcode: int) -> None:
            items = connectivity.GetNetItems(
                netcode, [pcbnew.PCB_TRACE_T, pcbnew.PCB_VIA_T]
            )
            for item in items:
                item_copy = item.Duplicate()
                item_copy.SetNetCode(0)
                if KICAD_VERSION >= (7, 0, 0):
                    item_copy.Move(pcbnew.VECTOR2I(-origin.x, -origin.y))
                else:
                    item_copy.Move(pcbnew.wxPoint(-origin.x, -origin.y))

                self.board.RemoveNative(item)
                result.append(item_copy)

        for net in common_nets:
            _append_normalized_connection_items(net)

        # some switch footprints (for example reversible) can have multiple pads
        # with the same netcode which might 'pre-routed' as well.
        # get nets unique to switch (not shared with diode) and collect connection
        # temple from them as well.
        switch_nets = [p.GetNetCode() for p in switch.Pads()]
        switch_unique_nets = list(set(switch_nets) - set(common_nets))
        for net in switch_unique_nets:
            _append_normalized_connection_items(net)

        def _format_item(item: pcbnew.PCB_TRACK) -> str:
            if KICAD_VERSION >= (7, 0, 0):
                name = item.GetFriendlyName()
            else:
                name = "Via" if item.Type() == pcbnew.PCB_VIA_T else "Track"
            start = item.GetStart()
            end = item.GetEnd()
            return f"{name} [{start} {end}]"

        items_str = ", ".join([_format_item(i) for i in result])
        logging.info(f"Got connection template: {items_str}")

        if destination_path:
            self.save_connection_template(switch, diode, result, destination_path)

        return result

    def load_connection_preset(
        self, key_format: str, diode_format: str, source_path: str
    ) -> List[pcbnew.PCB_TRACK]:
        board = pcbnew.LoadBoard(source_path)
        tracks = board.GetTracks()
        for t in tracks:
            t.SetNetCode(0)
        return tracks

    def place_switches(
        self,
        keyboard: Keyboard,
        key_format: str,
    ) -> None:
        current_key = 1

        for key in keyboard.keys:
            switch_footprint = get_footprint(self.board, key_format.format(current_key))

            width = key.width
            height = key.height
            position = (
                pcbnew.wxPoint(
                    (self.__key_distance_x * key.x)
                    + (self.__key_distance_x * width // 2),
                    (self.__key_distance_y * key.y)
                    + (self.__key_distance_y * height // 2),
                )
                + self.__reference_coordinate
            )
            set_position(switch_footprint, position)
            reset_rotation(switch_footprint)

            angle = key.rotation_angle
            if angle != 0:
                rotation_reference = (
                    pcbnew.wxPoint(
                        (self.__key_distance_x * key.rotation_x),
                        (self.__key_distance_y * key.rotation_y),
                    )
                    + self.__reference_coordinate
                )
                rotate(switch_footprint, rotation_reference, angle)

            current_key += 1

    def place_switch_elements(
        self,
        key_format: str,
        elements: List[ElementInfo],
    ) -> None:
        for i, switch_footprint in SwitchIterator(self.board, key_format):
            position = get_position(switch_footprint)
            orientation = get_orientation(switch_footprint)
            for element_info in elements:
                annotation_format = element_info.annotation_format
                element_position = element_info.position
                footprint = get_optional_footprint(
                    self.board, annotation_format.format(i)
                )
                if footprint and element_position:
                    reset_rotation(footprint)
                    set_side(footprint, element_position.side)
                    set_rotation(footprint, element_position.orientation)

                    offset = pcbnew.wxPointMM(
                        *element_position.relative_position.to_list()
                    )
                    if orientation != 0:
                        offset = position_in_rotated_coordinates(offset, orientation)

                    set_position(footprint, position + offset)
                    if orientation != 0:
                        current_position = get_position(footprint)
                        rotate(footprint, current_position, -1 * orientation)

    def route_switches_with_diodes(
        self,
        key_format: str,
        diode_format: str,
        template_connection: List[pcbnew.PCB_TRACK],
    ) -> None:
        for i, switch_footprint in SwitchIterator(self.board, key_format):
            angle = -1 * switch_footprint.GetOrientationDegrees()

            if diode_footprint := get_optional_footprint(
                self.board, diode_format.format(i)
            ):
                self.route_switch_with_diode(
                    switch_footprint, diode_footprint, angle, template_connection
                )
        # when done, delete all template items
        for item in template_connection:
            self.board.RemoveNative(item)

    def route_rows_and_columns(self) -> None:
        column_pads = {}
        row_pads = {}

        sorted_pads = pcbnew.PADS_VEC()
        self.board.GetSortedPadListByXthenYCoord(sorted_pads)
        for pad in sorted_pads:
            net_name = pad.GetNetname()
            if match := re.match(r"^COL(\d+)$", net_name, re.IGNORECASE):
                column_number = match.groups()[0]
                column_pads.setdefault(column_number, []).append(pad)
            elif match := re.match(r"^ROW(\d+)$", net_name, re.IGNORECASE):
                row_number = match.groups()[0]
                row_pads.setdefault(row_number, []).append(pad)

        # very naive routing approach, will fail in some scenarios:
        for pads in column_pads.values():
            for pad1, pad2 in zip(pads, pads[1:]):
                if pad1.GetParentAsString() == pad2.GetParentAsString():
                    # do not connect pads of the same footprint
                    continue
                self.route(pad1, pad2)

        for pads in row_pads.values():
            for pad1, pad2 in zip(pads, pads[1:]):
                self.route(pad1, pad2)

    def run(
        self,
        layout: dict,
        key_format: str,
        diode_info: Optional[ElementInfo],
        route_switches_with_diodes: bool = False,
        route_rows_and_columns: bool = False,
        additional_elements: List[ElementInfo] = [],
    ) -> None:
        diode_format = ""
        template_connection = []

        def _normalize_template_path(template_path: str) -> str:
            if str(Path(template_path).name) == template_path:
                # if passed filename without directory,
                # assume that this refers to file in the directory of
                # current board. This is mostly for CLI convenience
                template_path = str(
                    Path(self.board.GetFileName()).parent / template_path
                )
            return template_path

        if diode_info:
            logging.info(f"Diode info: {diode_info}")
            diode_format = diode_info.annotation_format
            if diode_info.position_option == PositionOption.RELATIVE:
                template_connection = self.get_connection_template(
                    key_format, diode_format, diode_info.template_path
                )
            elif diode_info.position_option == PositionOption.PRESET:
                logging.info(
                    f"Loading diode connection preset from {diode_info.template_path}"
                )
                template_connection = self.load_connection_preset(
                    key_format,
                    diode_format,
                    _normalize_template_path(diode_info.template_path),
                )
            additional_elements = [diode_info] + additional_elements

        for element_info in additional_elements:
            if element_info.position_option in [
                PositionOption.RELATIVE,
                PositionOption.PRESET,
            ]:
                if element_info.position_option == PositionOption.PRESET:
                    source = pcbnew.IO_MGR.Load(
                        pcbnew.IO_MGR.KICAD_SEXP,
                        _normalize_template_path(element_info.template_path),
                    )
                else:
                    source = self.board

                element1 = get_footprint(source, key_format.format(1))
                element2 = get_footprint(
                    source, element_info.annotation_format.format(1)
                )
                position = self.get_current_relative_element_position(
                    element1, element2
                )
                element_info.position = position
                logging.info(f"Element info updated: {element_info}")

        if layout:
            logging.info(f"User layout: {layout}")
            keyboard = get_keyboard(layout)
            self.place_switches(keyboard, key_format)

        if additional_elements:
            self.place_switch_elements(key_format, additional_elements)

        if route_switches_with_diodes:
            self.route_switches_with_diodes(
                key_format, diode_format, template_connection
            )

        if route_rows_and_columns:
            self.route_rows_and_columns()

        if route_switches_with_diodes or route_rows_and_columns:
            self.remove_dangling_tracks()



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, Flag
from typing import Optional


class Side(Flag):
    FRONT = False
    BACK = True

    @classmethod
    def get(cls, name) -> Side:
        if isinstance(name, str):
            if name.upper() == "FRONT":
                return Side.FRONT
            elif name.upper() == "BACK":
                return Side.BACK
        msg = f"'{name}' is not valid Position"
        raise ValueError(msg)


@dataclass
class Point:
    x: float
    y: float

    def to_list(self) -> list[float]:
        return [self.x, self.y]


@dataclass
class ElementPosition:
    relative_position: Point
    orientation: float
    side: Side


class PositionOption(str, Enum):
    DEFAULT = "Default"
    RELATIVE = "Relative"
    PRESET = "Preset"
    CUSTOM = "Custom"

    def __str__(self) -> str:
        return self.value

    @classmethod
    def get(cls, name) -> PositionOption:
        if isinstance(name, str):
            value = name.replace("_", " ").title()
            return PositionOption(value)
        msg = f"'{name}' is not valid Position"
        raise ValueError(msg)


@dataclass
class ElementInfo:
    annotation_format: str
    position_option: PositionOption
    position: Optional[ElementPosition]
    template_path: str

    def to_dict(self) -> dict:
        value = {
            "annotation_format": self.annotation_format,
            "position_option": self.position_option,
            "position": {
                "relative_position": self.position.relative_position.to_list(),
                "orientation": self.position.orientation,
                "side": self.position.side.name,
            }
            if self.position
            else None,
            "template_path": self.template_path,
        }
        return value

    @classmethod
    def from_dict(cls, value: dict) -> ElementInfo:
        try:
            annotation_format = value["annotation_format"]
            position_option = PositionOption.get(value["position_option"])
            if position := value["position"]:
                x, y = position["relative_position"]
                orientation = position["orientation"]
                side = Side.get(position["side"])
                position = ElementPosition(Point(x, y), orientation, side)
            else:
                position = None
            template_path = value["template_path"]
            return ElementInfo(
                annotation_format, position_option, position, template_path
            )
        except Exception as e:
            msg = "Failed to create ElementInfo object"
            raise ValueError(msg) from e



import inspect


def __is_in_call_stack(function_name: str, module_name: str) -> bool:
    current_stack = inspect.stack()

    for frame_info in current_stack:
        frame = frame_info.frame
        if frame.f_globals.get("__name__") == module_name:
            if function_name in frame.f_locals or function_name in frame.f_globals:
                return True

    return False


if __is_in_call_stack("LoadPluginModule", "pcbnew"):
    from .kbplacer_plugin_action import KbplacerPluginAction

    KbplacerPluginAction().register()



from __future__ import annotations

import builtins
import itertools
import logging
import math
from typing import Tuple

import pcbnew

from .element_position import Side

# remove pre-release and build numbers (if present) and split to major-minor-patch tuple
KICAD_VERSION = tuple(
    map(int, ((pcbnew.Version().split("+")[0]).split("-")[0]).split("."))
)
DEFAULT_CLEARANCE_MM = 0.25


def position_in_rotated_coordinates(
    point: pcbnew.wxPoint, angle: float
) -> pcbnew.wxPoint:
    """
    Map position in xy-Cartesian coordinate system to x'y'-Cartesian which
    has same origin but axes are rotated by angle

    :param point: A point to be mapped
    :param angle: Rotation angle (in degrees) of x'y'-Cartesian coordinates
    :type point: pcbnew.wxPoint
    :type angle: float
    :return: Result position in x'y'-Cartesian coordinates
    :rtype: pcbnew.wxPoint
    """
    x, y = point.x, point.y
    angle = math.radians(angle)
    xr = (x * math.cos(angle)) + (y * math.sin(angle))
    yr = (-x * math.sin(angle)) + (y * math.cos(angle))
    return pcbnew.wxPoint(xr, yr)


def position_in_cartesian_coordinates(
    point: pcbnew.wxPoint, angle: float
) -> pcbnew.wxPoint:
    """Performs inverse operation to position_in_rotated_coordinates i.e.
    map position in rotated (by angle) x'y'-Cartesian to xy-Cartesian

    :param point: A point to be mapped
    :param angle: Rotation angle (in degrees) of x'y'-Cartesian coordinates
    :type point: pcbnew.wxPoint
    :type angle: float
    :return: Result position in xy-Cartesian coordinates
    :rtype: pcbnew.wxPoint
    """
    xr, yr = point.x, point.y
    angle = math.radians(angle)
    x = (xr * math.cos(angle)) - (yr * math.sin(angle))
    y = (xr * math.sin(angle)) + (yr * math.cos(angle))
    return pcbnew.wxPoint(x, y)


def get_footprint(board: pcbnew.BOARD, reference: str) -> pcbnew.FOOTPRINT:
    logging.info(f"Searching for {reference} footprint in {board.GetFileName()}")
    footprint = board.FindFootprintByReference(reference)
    if footprint is None:
        logging.error("Footprint not found")
        msg = f"Cannot find footprint {reference}"
        raise Exception(msg)
    return footprint


def get_optional_footprint(
    board: pcbnew.BOARD, reference: str
) -> pcbnew.FOOTPRINT | None:
    try:
        footprint = get_footprint(board, reference)
    except Exception as _:
        footprint = None
    return footprint


def set_position(footprint: pcbnew.FOOTPRINT, position: pcbnew.wxPoint) -> None:
    logging.debug(f"Setting {footprint.GetReference()} footprint position: {position}")
    if KICAD_VERSION >= (7, 0, 0):
        footprint.SetPosition(pcbnew.VECTOR2I(position.x, position.y))
    else:
        footprint.SetPosition(position)


def set_position_by_points(footprint: pcbnew.FOOTPRINT, x: int, y: int) -> None:
    set_position(footprint, pcbnew.wxPoint(x, y))


def get_position(footprint: pcbnew.FOOTPRINT) -> pcbnew.wxPoint:
    position = footprint.GetPosition()
    logging.debug(f"Getting {footprint.GetReference()} footprint position: {position}")
    if KICAD_VERSION >= (7, 0, 0):
        return pcbnew.wxPoint(position.x, position.y)
    return position


def set_side(footprint: pcbnew.FOOTPRINT, side: Side) -> None:
    if side ^ get_side(footprint):
        footprint.Flip(footprint.GetPosition(), False)


def get_side(footprint: pcbnew.FOOTPRINT) -> Side:
    return Side(footprint.IsFlipped())


def set_rotation(footprint: pcbnew.FOOTPRINT, angle: float) -> None:
    footprint.SetOrientationDegrees(angle)


def reset_rotation(footprint: pcbnew.FOOTPRINT) -> None:
    set_rotation(footprint, 0)


def get_orientation(footprint: pcbnew.FOOTPRINT) -> float:
    return footprint.GetOrientationDegrees()


def rotate(
    item: pcbnew.BOARD_ITEM,
    rotation_reference: pcbnew.wxPoint,
    angle: float,
) -> None:
    if KICAD_VERSION >= (7, 0, 0):
        item.Rotate(
            pcbnew.VECTOR2I(rotation_reference.x, rotation_reference.y),
            pcbnew.EDA_ANGLE(angle * -1, pcbnew.DEGREES_T),
        )
    else:
        item.Rotate(rotation_reference, angle * -10)


def get_distance(i1: pcbnew.BOARD_ITEM, i2: pcbnew.BOARD_ITEM) -> int:
    """Calculate distance between two board items"""
    center1 = i1.GetPosition()
    center2 = i2.GetPosition()
    return int(((center1.x - center2.x) ** 2 + (center1.y - center2.y) ** 2) ** 0.5)


def get_common_layers(p1: pcbnew.PAD, p2: pcbnew.PAD) -> list[int]:
    """Returns list of common layer ids for both of the given pads,
    may be empty if no common layers found
    """
    set1 = [layer for layer in p1.GetLayerSet().CuStack()]
    set2 = [layer for layer in p2.GetLayerSet().CuStack()]
    return list(set(set1).intersection(set2))


def get_common_nets(f1: pcbnew.FOOTPRINT, f2: pcbnew.FOOTPRINT) -> list[int]:
    """Returns list of netcodes which are used by both of the
    given footprints, may be empty if no common nets found
    """
    codes1 = [p.GetNetCode() for p in f1.Pads()]
    codes2 = [p.GetNetCode() for p in f2.Pads()]
    return list(set(codes1).intersection(codes2))


def get_closest(
    pads1: list[pcbnew.PAD], pads2: list[pcbnew.PAD]
) -> Tuple[int, Tuple[pcbnew.PAD, pcbnew.PAD]]:
    all_pairs = itertools.product(pads1, pads2)
    min_distance, closest_pair = min(
        ((get_distance(obj1, obj2), (obj1, obj2)) for obj1, obj2 in all_pairs),
        key=lambda x: x[0],
    )
    return min_distance, closest_pair


def get_closest_pads_on_same_net(
    f1: pcbnew.FOOTPRINT, f2: pcbnew.FOOTPRINT
) -> Tuple[pcbnew.PAD, pcbnew.PAD] | None:
    """Return pair of closest, same net pads of two given footprints
    or None if footprints do not use common net
    """
    pads1 = {}
    for p in f1.Pads():
        if netname := p.GetNetname():
            pads1.setdefault(netname, []).append(p)

    pads2 = {}
    for p in f2.Pads():
        netname = p.GetNetname()
        if netname in pads1:
            pads2.setdefault(netname, []).append(p)

    result = None
    closest = float("inf")
    for key, value in pads2.items():
        distance, result_temp = get_closest(value, pads1[key])
        if distance < closest:
            closest = distance
            result = result_temp

    # We iterated with pads2 so we need to swap reult to keep original
    # order in tuple, first pad should be from f1, second from f2
    # Some caller might require proper order
    if result:
        return result[1], result[0]
    return result


class BoardModifier:
    def __init__(self, board: pcbnew.BOARD) -> None:
        self.board = board

    def get_connectivity(self):
        self.board.BuildConnectivity()
        return self.board.GetConnectivity()

    def test_track_collision(self, track: pcbnew.PCB_TRACK) -> bool:
        collide_list = []
        track_shape = track.GetEffectiveShape()
        track_start = track.GetStart()
        track_end = track.GetEnd()
        track_net_code = track.GetNetCode()
        # connectivity needs to be last,
        # otherwise it will update track net name before we want it to:
        connectivity = self.get_connectivity()
        for f in self.board.GetFootprints():
            reference = f.GetReference()
            hull = f.GetBoundingHull()
            if hit_test_result := hull.Collide(track_shape):
                for p in f.Pads():
                    pad_name = p.GetName()
                    pad_shape = p.GetEffectiveShape()

                    # track has non default netlist set so we can skip
                    # collision detection for pad of same netlist:
                    if track_net_code != 0 and track_net_code == p.GetNetCode():
                        logging.debug(
                            f"Track collision ignored, pad {reference}:{pad_name} "
                            f"on same netlist: {track.GetNetname()}/{p.GetNetname()}"
                        )
                        continue

                    # if track starts or ends in pad then assume that
                    # this collision is expected, with the exception of case
                    # where track already has netlist set and it is different
                    # than pad's netlist or pad has `NPTH` attribute
                    if p.HitTest(track_start) or p.HitTest(track_end):
                        if (
                            track_net_code != 0
                            and track_net_code != p.GetNetCode()
                            and p.IsOnLayer(track.GetLayer())
                        ) or p.GetAttribute() == pcbnew.PAD_ATTRIB_NPTH:
                            logging.debug(
                                f"Track collide with pad {reference}:{pad_name}"
                            )
                            collide_list.append(p)
                        else:
                            logging.debug(
                                "Track collision ignored, track starts or ends "
                                f"in pad {reference}:{pad_name}"
                            )
                    else:
                        hit_test_result = pad_shape.Collide(
                            track_shape, pcbnew.FromMM(DEFAULT_CLEARANCE_MM)
                        )
                        on_same_layer = p.IsOnLayer(track.GetLayer())
                        if hit_test_result and on_same_layer:
                            logging.debug(
                                f"Track collide with pad {reference}:{pad_name}"
                            )
                            collide_list.append(p)
        # track ids to clear at the end:
        tracks_to_clear = []
        for t in self.board.GetTracks():
            # check collision if not itself, on same layer and with different netlist
            # (unless 'trackNetCode' is default '0' netlist):
            if (
                t.m_Uuid != track.m_Uuid
                and t.IsOnLayer(track.GetLayer())
                and (track_net_code != t.GetNetCode() or track_net_code == 0)
            ):
                track_uuid = t.m_Uuid.AsString()
                if (
                    track_start == t.GetStart()
                    or track_start == t.GetEnd()
                    or track_end == t.GetStart()
                    or track_end == t.GetEnd()
                ):
                    logging.debug(
                        "Track collision ignored, track starts or ends "
                        f"at the end of {track_uuid} track"
                    )
                    # ignoring one track means that we can ignore
                    # all other connected to it:
                    tracks_to_clear += [
                        x.m_Uuid for x in connectivity.GetConnectedTracks(t)
                    ]
                    # check if connection to this track clears pad collision:
                    connected_pads_ids = [
                        x.m_Uuid for x in connectivity.GetConnectedPads(t)
                    ]
                    for collision in list(collide_list):
                        if collision.m_Uuid in connected_pads_ids:
                            logging.debug(
                                "Pad collision removed due to connection with track "
                                "which leads to that pad"
                            )
                            collide_list.remove(collision)
                elif hit_test_result := t.GetEffectiveShape().Collide(
                    track_shape, pcbnew.FromMM(DEFAULT_CLEARANCE_MM)
                ):
                    logging.debug(f"Track collide with another track: {track_uuid}")
                    collide_list.append(t)
        for collision in list(collide_list):
            if collision.m_Uuid in tracks_to_clear:
                collision_uuid = collision.m_Uuid.AsString()
                logging.debug(
                    f"Track collision with {collision_uuid} removed due to "
                    "connection with track which leads to it"
                )
                collide_list.remove(collision)
        return len(collide_list) != 0

    def add_track_to_board(self, track: pcbnew.PCB_TRACK):
        """Add track to the board if track passes collision check.
        If track has no set netlist, it would get netlist of a pad
        or other track, on which it started or ended.
        Collision with element of the same netlist will be ignored
        unless it is default '0' netlist.
        This exception about '0' netlist is important because it helps
        to detect collisions with holes.

        :param track: A track to be added to board
        :return: End position of added track or None if failed to add.
        """
        if not self.test_track_collision(track):
            layer_name = self.board.GetLayerName(track.GetLayer())
            start = track.GetStart()
            stop = track.GetEnd()
            logging.info(
                f"Adding track segment ({layer_name}): [{start}, {stop}]",
            )
            self.board.Add(track)
            return stop
        else:
            logging.warning("Could not add track segment due to detected collision")
            return None

    def add_track_segment_by_points(
        self, start: pcbnew.wxPoint, end: pcbnew.wxPoint, layer=pcbnew.B_Cu
    ):
        track = pcbnew.PCB_TRACK(self.board)
        track.SetWidth(pcbnew.FromMM(0.25))
        track.SetLayer(layer)
        if KICAD_VERSION >= (7, 0, 0):
            track.SetStart(pcbnew.VECTOR2I(start.x, start.y))
            track.SetEnd(pcbnew.VECTOR2I(end.x, end.y))
        else:
            track.SetStart(start)
            track.SetEnd(end)
        return self.add_track_to_board(track)

    def add_track_segment(
        self, start: pcbnew.wxPoint, vector: list[int], layer=pcbnew.B_Cu
    ):
        end = pcbnew.wxPoint(start.x + vector[0], start.y + vector[1])
        return self.add_track_segment_by_points(start, end, layer)

    def route(self, pad1: pcbnew.PAD, pad2: pcbnew.PAD) -> None:
        r"""Simple track router
        If pads are collinear, it will use single track segment.
        If pads are not collinear it will try use two segments track,
        in first try, it will fanout straight segment from `pad1` to
        `pad2` direction and then it will end it with second segment
        at 45degree angle:

        |pad| ____
        | 1 |     \
                   \
                  |pad|
                  | 2 |

        If this does not succeed (it is not possible using 45degree angles
        and two segments only or there is collision detected),
        then it will try to switch track posture:

        |pad|
        | 1 |
          \
           \______|pad|
                  | 2 |

        If this is not possible as well, it will leave pads unconnected
        and will not try another route.

        If both parent footprint are rotated the same, it might
        route track at that angle.
        Footprint is considered rotated, if its orientation is different than
        0, 90, 180 and 270. When comparing rotations use angle to closest
        quorter, i.e. footprints with orientation 10 and 190 (190-180=10) are
        considered to be rotated the same.
        """
        layers = get_common_layers(pad1, pad2)
        if not layers:
            logging.warning("Could not route pads, no common layers found")
            return

        layer = layers[0]
        logging.debug(f"Routing at {self.board.GetLayerName(layer)} layer")

        def _calculate_corners(
            pos1: pcbnew.wxPoint, pos2: pcbnew.wxPoint
        ) -> Tuple[pcbnew.wxPoint, pcbnew.wxPoint]:
            x_diff = pos1.x - pos2.x
            y_diff = pos1.y - pos2.y
            x_diff_abs = builtins.abs(x_diff)
            y_diff_abs = builtins.abs(y_diff)
            if x_diff_abs < y_diff_abs:
                up_or_down = -1 if y_diff > 0 else 1
                return (
                    pcbnew.wxPoint(
                        pos2.x + x_diff,
                        pos2.y - (up_or_down * x_diff_abs),
                    ),
                    pcbnew.wxPoint(pos1.x - x_diff, pos1.y - x_diff),
                )
            else:
                left_or_right = -1 if x_diff > 0 else 1
                return (
                    pcbnew.wxPoint(
                        pos2.x - (left_or_right * y_diff_abs),
                        pos2.y + y_diff,
                    ),
                    pcbnew.wxPoint(pos1.x - y_diff, pos1.y - y_diff),
                )

        def _route(
            pos1: pcbnew.wxPoint, pos2: pcbnew.wxPoint, corner: pcbnew.wxPoint
        ) -> bool:
            if end := self.add_track_segment_by_points(pos1, corner, layer):
                end = self.add_track_segment_by_points(end, pos2, layer)
                return end is not None
            return False

        pos1 = pad1.GetPosition()
        pos2 = pad2.GetPosition()

        orientation1 = get_orientation(pad1.GetParent())
        orientation2 = get_orientation(pad2.GetParent())

        if (orientation1 % 90 == 0) and (orientation2 % 90 == 0):
            angle = 0
        elif orientation1 % 90 == orientation2 % 90:
            # if rotations are considered the same use the angle
            angle = (-1 * orientation1 + 360) % 360
        else:
            logging.warning(
                "Could not route pads when parent footprints not rotated the same"
            )
            return

        logging.debug(f"Routing pad {pos1} with pad {pos2} at {angle} degree angle")

        # if in line, use one track segment
        if pos1.x == pos2.x or pos1.y == pos2.y:
            self.add_track_segment_by_points(pos1, pos2, layer)
        else:
            # pads are not in single line, attempt routing with two segment track
            if angle != 0:
                pos1_r = position_in_rotated_coordinates(pos1, angle)
                pos2_r = position_in_rotated_coordinates(pos2, angle)
                corners = _calculate_corners(pos1_r, pos2_r)
                corners = (
                    position_in_cartesian_coordinates(corners[0], angle),
                    position_in_cartesian_coordinates(corners[1], angle),
                )
            else:
                corners = _calculate_corners(pos1, pos2)

            if not _route(pos1, pos2, corners[0]):
                _route(pos1, pos2, corners[1])



from __future__ import annotations

import argparse
import copy
import json
import pprint
import sys
from dataclasses import asdict, dataclass, field, fields
from typing import Any, List, Optional, Tuple, Type, Union

DEFAULT_KEY_COLOR = "#cccccc"
DEFAULT_TEXT_COLOR = "#000000"
DEFAULT_TEXT_SIZE = 3

# Map from serialized label position to normalized position,
# depending on the alignment flags.
# fmt: off
LABEL_MAP: List[List[int]] = [
    # 0  1  2  3  4  5  6  7  8  9 10 11   # align flags
    [ 0, 6, 2, 8, 9,11, 3, 5, 1, 4, 7,10], # 0 = no centering
    [ 1, 7,-1,-1, 9,11, 4,-1,-1,-1,-1,10], # 1 = center x
    [ 3,-1, 5,-1, 9,11,-1,-1, 4,-1,-1,10], # 2 = center y
    [ 4,-1,-1,-1, 9,11,-1,-1,-1,-1,-1,10], # 3 = center x & y
    [ 0, 6, 2, 8,10,-1, 3, 5, 1, 4, 7,-1], # 4 = center front (default)
    [ 1, 7,-1,-1,10,-1, 4,-1,-1,-1,-1,-1], # 5 = center front & x
    [ 3,-1, 5,-1,10,-1,-1,-1, 4,-1,-1,-1], # 6 = center front & y
    [ 4,-1,-1,-1,10,-1,-1,-1,-1,-1,-1,-1], # 7 = center front & x & y
]

REVERSE_LABEL_MAP: List[List[int]] = [
    # 0  1  2  3  4  5  6  7  8  9 10 11   # align flags
    [ 0, 8, 2, 6, 9, 7, 1,10, 3, 4,11, 5], # 0 = no centering
    [-1, 0,-1,-1, 6,-1,-1, 1,-1, 4,11, 5], # 1 = center x
    [-1,-1,-1, 0, 8, 2,-1,-1,-1, 4,11, 5], # 2 = center y
    [-1,-1,-1,-1, 0,-1,-1,-1,-1, 4,11, 5], # 3 = center x & y
    [ 0, 8, 2, 6, 9, 7, 1,10, 3,-1, 4,-1], # 4 = center front (default)
    [-1, 0,-1,-1, 6,-1,-1, 1,-1,-1, 4,-1], # 5 = center front & x
    [-1,-1,-1, 0, 8, 2,-1,-1,-1,-1, 4,-1], # 6 = center front & y
    [-1,-1,-1,-1, 0,-1,-1,-1,-1,-1, 4,-1], # 7 = center front & x & y
]
# fmt: on


@dataclass
class KeyDefault:
    textColor: str = DEFAULT_TEXT_COLOR
    textSize: int = DEFAULT_TEXT_SIZE


@dataclass
class Key:
    color: str = DEFAULT_KEY_COLOR
    labels: List[str] = field(default_factory=list)
    textColor: List[Optional[str]] = field(default_factory=list)
    textSize: List[Optional[int]] = field(default_factory=list)
    default: KeyDefault = field(default_factory=KeyDefault)
    x: float = 0
    y: float = 0
    width: float = 1
    height: float = 1
    x2: float = 0
    y2: float = 0
    width2: float = 1
    height2: float = 1
    rotation_x: float = 0
    rotation_y: float = 0
    rotation_angle: float = 0
    decal: bool = False
    ghost: bool = False
    stepped: bool = False
    nub: bool = False
    profile: str = ""
    sm: str = ""  # switch mount
    sb: str = ""  # switch brand
    st: str = ""  # switch type

    def __post_init__(self: Key) -> None:
        if isinstance(self.default, dict):
            self.default = KeyDefault(**self.default)
        for field in self.__dataclass_fields__:
            value = getattr(self, field)
            if isinstance(value, float):
                new_val = round(value, 6)
                setattr(self, field, new_val)


@dataclass
class Background:
    name: str = ""
    style: str = ""


@dataclass
class KeyboardMetadata:
    author: str = ""
    backcolor: str = "#eeeeee"
    background: Optional[Background] = None
    name: str = ""
    notes: str = ""
    radii: str = ""
    switchBrand: str = ""
    switchMount: str = ""
    switchType: str = ""

    def __post_init__(self: KeyboardMetadata) -> None:
        if isinstance(self.background, dict):
            self.background = Background(**self.background)


@dataclass
class Keyboard:
    meta: KeyboardMetadata = field(default_factory=KeyboardMetadata)
    keys: List[Key] = field(default_factory=list)

    @classmethod
    def from_json(cls: Type[Keyboard], data: dict) -> Keyboard:
        if isinstance(data["meta"], dict):
            data["meta"] = KeyboardMetadata(**data["meta"])
        if isinstance(data["keys"], list):
            keys: List[Key] = [Key(**key) for key in data["keys"]]
            data["keys"] = keys
        return cls(**data)

    def to_json(self: Keyboard, indent: Optional[int] = None) -> str:
        return json.dumps(asdict(self), indent=indent)

    def __text_size_changed(self: Keyboard, current: list[Any], new: list[Any]) -> bool:
        current = copy.copy(current)
        new = copy.copy(new)
        for obj in [current, new]:
            if len_difference := 12 - len(obj):
                obj.extend(len_difference * [0])
        return current != new

    def to_kle(self: Keyboard) -> str:
        row = []
        rows = []

        current: Key = copy.deepcopy(Key())
        # some properties are not part of Key type, store them separately:
        current_alignment = 4
        current_f2 = -1
        # rotation origin:
        cluster: dict[str, float] = {"r": 0, "rx": 0, "ry": 0}

        new_row = True
        current.y -= 1  # will be incremented on first row

        for key in self.keys:
            props: dict[str, Any] = {}

            def add_prop(name: str, value: Any, default: Any) -> Any:
                if isinstance(value, float):
                    value = round(value, 6)
                if value != default:
                    props[name] = value
                return value

            if key.labels:
                alignment, labels = find_best_label_alignment(key.labels)
            else:
                alignment, labels = 7, []

            # detect new row
            new_cluster = (
                key.rotation_angle != cluster["r"]
                or key.rotation_x != cluster["rx"]
                or key.rotation_y != cluster["ry"]
            )
            new_row = key.y != current.y
            if row and (new_cluster or new_row):
                # push the old row
                rows.append(row)
                row = []
                new_row = True

            if new_row:
                current.y = round(current.y + 1, 6)
                # 'y' is reset if either 'rx' or 'ry' are changed
                if key.rotation_y != cluster["ry"] or key.rotation_x != cluster["rx"]:
                    current.y = key.rotation_y
                # always reset x to rx (which defaults to zero)
                current.x = key.rotation_x

                cluster["r"] = key.rotation_angle
                cluster["rx"] = key.rotation_x
                cluster["ry"] = key.rotation_y

                new_row = False

            current.rotation_angle = add_prop(
                "r", key.rotation_angle, current.rotation_angle
            )
            current.rotation_x = add_prop("rx", key.rotation_x, current.rotation_x)
            current.rotation_y = add_prop("ry", key.rotation_y, current.rotation_y)

            x_offset = add_prop("x", round(key.x - current.x, 6), 0)
            y_offset = add_prop("y", round(key.y - current.y, 6), 0)
            current.x = round(current.x + key.width + x_offset, 6)
            current.y = round(current.y + y_offset, 6)

            current.color = add_prop("c", key.color, current.color)
            if text_color := reorder_items_kle(key.textColor, alignment):
                if not text_color[0]:
                    text_color[0] = key.default.textColor
                text_color = ["" if not item else item for item in text_color]
                text_color = "\n".join(text_color).rstrip("\n")
                current.textColor = add_prop("t", text_color, current.textColor)
            else:
                current.default.textColor = add_prop(
                    "t", key.default.textColor, current.default.textColor
                )

            current.ghost = add_prop("g", key.ghost, current.ghost)
            current.profile = add_prop("p", key.profile, current.profile)
            current.sm = add_prop("sm", key.sm, current.sm)
            current.sb = add_prop("sb", key.sb, current.sb)
            current.st = add_prop("st", key.st, current.st)

            current_alignment = add_prop("a", alignment, current_alignment)
            current.default.textSize = add_prop(
                "f", key.default.textSize, current.default.textSize
            )
            if "f" in props:
                current.textSize = []

            text_size = reorder_items_kle(key.textSize, alignment)
            text_size = [0 if not isinstance(i, int) else i for i in text_size]
            if self.__text_size_changed(current.textSize, text_size):
                if not text_size:
                    current.default.textSize = add_prop(
                        "f", key.default.textSize, current.default.textSize
                    )
                    current.textSize = []
                else:
                    if optimize := not text_size[0]:
                        optimize = all(x == text_size[1] for x in text_size[2:])
                    if optimize:
                        f2 = text_size[1]
                        current_f2 = add_prop("f2", f2, current_f2)
                        # don't know why this gives type checking error, works fine:
                        current.textSize = [0] + (11 * [f2])  # type: ignore
                    else:
                        current.textSize = add_prop("fa", text_size, [])

            add_prop("w", key.width, 1)
            add_prop("h", key.height, 1)
            add_prop("w2", key.width2, key.width)
            add_prop("h2", key.height2, key.height)
            add_prop("x2", key.x2, 0)
            add_prop("y2", key.y2, 0)
            add_prop("l", key.stepped, False)
            add_prop("n", key.nub, False)
            add_prop("d", key.decal, False)

            if props:
                row.append(props)

            current.labels = labels
            labels = ["" if not item else item for item in labels]
            row.append("\n".join(labels).rstrip("\n"))

        if row:
            rows.append(row)

        result = ""

        default_meta = asdict(KeyboardMetadata())
        meta = copy.deepcopy(asdict(self.meta))
        if meta != default_meta:
            # include only non-default meta fields
            for k in list(meta.keys()):
                if default_meta.get(k, None) == meta[k]:
                    del meta[k]
            result += json.dumps(meta, indent=None) + ",\n"

        for row in rows:
            result += json.dumps(row, indent=None) + ",\n"
        result = result.strip(",\n")
        return result


@dataclass
class ViaKeyboard(Keyboard):
    alternative_keys: List[Key] = field(default_factory=list)

    LAYOUT_LABEL = 8

    def __post_init__(self: ViaKeyboard) -> None:
        for key in list(self.keys):
            if self.is_alternative(key):
                self.alternative_keys.append(copy.deepcopy(key))
                self.keys.remove(key)

    def is_alternative(self, key: Key) -> bool:
        labels = key.labels
        if key.decal or len(labels) >= 9 and labels[self.LAYOUT_LABEL]:
            # check if not default layout:
            if labels[self.LAYOUT_LABEL].split(",")[1] != "0":
                # alternative layout key
                return True

        return False


def reorder_items(items: List[Any], align: int) -> List[Any]:
    ret: List[Any] = 12 * [None]
    for i, item in enumerate(items):
        if item:
            index = LABEL_MAP[align][i]
            ret[index] = item
    while ret and ret[-1] is None:
        ret.pop()
    return ret


def reorder_items_kle(items, align) -> List[Any]:
    ret: List[Any] = 12 * [None]
    for i, label in enumerate(items):
        if label:
            index = REVERSE_LABEL_MAP[align][i]
            if index == -1:
                ret = []
                break
            ret[index] = label
    while ret and ret[-1] is None:
        ret.pop()
    return ret


def find_best_label_alignment(labels) -> Tuple[int, List[Any]]:
    results = {}
    for align in reversed(range(0, 8)):
        if ret := reorder_items_kle(labels, align):
            results[align] = ret

    if results.items():
        best = min(results.items(), key=lambda x: len(x[1]))
        return best[0], best[1]
    else:
        return 0, []


def cleanup_key(key: Key):
    for attribute_name in ["textSize", "textColor"]:
        attribute = getattr(key, attribute_name)
        attribute = attribute[0 : len(key.labels)]
        for i, (label, val) in enumerate(zip(key.labels, attribute)):
            if not label:
                attribute[i] = None
            if val == getattr(key.default, attribute_name):
                attribute[i] = None
        while attribute and attribute[-1] is None:
            attribute.pop()
        setattr(key, attribute_name, attribute)


def parse_via(layout) -> ViaKeyboard:
    keyboard = parse_kle(layout["layouts"]["keymap"])
    return ViaKeyboard(meta=keyboard.meta, keys=keyboard.keys)


def parse_kle(layout) -> Keyboard:
    if not isinstance(layout, list):
        msg = "Expected an list of objects"
        raise RuntimeError(msg)

    metadata: KeyboardMetadata = KeyboardMetadata()
    rows: List[Any] = layout
    current: Key = Key()

    if not rows:
        msg = "Expected at least one row of keys"
        raise RuntimeError(msg)

    keys = []
    cluster = {"x": 0, "y": 0}
    align = 4

    for r, row in enumerate(rows):
        if isinstance(row, list):
            for k, item in enumerate(row):
                if isinstance(item, str):
                    new_key = copy.deepcopy(current)
                    # Calculate some generated values
                    new_key.width2 = (
                        current.width if new_key.width2 == 0 else current.width2
                    )
                    new_key.height2 = (
                        current.height if new_key.height2 == 0 else current.height2
                    )
                    new_key.labels = reorder_items(item.split("\n"), align)
                    new_key.textSize = reorder_items(new_key.textSize, align)

                    cleanup_key(new_key)

                    keys.append(new_key)

                    current.x = round(current.x + current.width, 6)
                    current.width = 1
                    current.height = 1
                    current.x2 = 0
                    current.y2 = 0
                    current.width2 = 0
                    current.height2 = 0
                    current.nub = False
                    current.stepped = False
                    current.decal = False
                elif isinstance(item, dict):
                    if k != 0 and ("r" in item or "rx" in item or "ry" in item):
                        msg = (
                            "Rotation can only be specified on the first key in the row"
                        )
                        raise RuntimeError(msg)
                    if "r" in item:
                        current.rotation_angle = item["r"]
                    if "rx" in item:
                        cluster["x"] = item["rx"]
                        current.x = cluster["x"]
                        current.y = cluster["y"]
                        current.rotation_x = item["rx"]
                    if "ry" in item:
                        cluster["y"] = item["ry"]
                        current.x = cluster["x"]
                        current.y = cluster["y"]
                        current.rotation_y = item["ry"]
                    if "a" in item:
                        align = item["a"]
                    if "f" in item:
                        current.default.textSize = item["f"]
                        current.textSize = []
                    if "f2" in item:
                        if len(current.textSize) == 0:
                            current.textSize = [None]
                        for _ in range(1, 12):
                            current.textSize.append(item["f2"])
                    if "fa" in item:
                        current.textSize = item["fa"]
                    if "p" in item:
                        current.profile = item["p"]
                    if "c" in item:
                        current.color = item["c"]
                    if "t" in item:
                        split = item["t"].split("\n")
                        if split[0]:
                            current.default.textColor = split[0]
                        current.textColor = reorder_items(split, align)
                    if "x" in item:
                        current.x = round(current.x + item["x"], 6)
                    if "y" in item:
                        current.y = round(current.y + item["y"], 6)
                    if "w" in item:
                        current.width = item["w"]
                        current.width2 = item["w"]
                    if "h" in item:
                        current.height = item["h"]
                        current.height2 = item["h"]
                    if "x2" in item:
                        current.x2 = item["x2"]
                    if "y2" in item:
                        current.y2 = item["y2"]
                    if "w2" in item:
                        current.width2 = item["w2"]
                    if "h2" in item:
                        current.height2 = item["h2"]
                    if "n" in item:
                        current.nub = item["n"]
                    if "l" in item:
                        current.stepped = item["l"]
                    if "d" in item:
                        current.decal = item["d"]
                    if "g" in item:
                        current.ghost = item["g"]
                    if "sm" in item:
                        current.sm = item["sm"]
                    if "sb" in item:
                        current.sb = item["sb"]
                    if "st" in item:
                        current.st = item["st"]
                else:
                    msg = "Unexpected item type"
                    raise RuntimeError(msg)

            # end of the row:
            current.y = round(current.y + 1, 6)
            current.x = current.rotation_x
        elif isinstance(row, dict) and r == 0:
            field_set = {f.name for f in fields(KeyboardMetadata) if f.init}
            row_filtered = {k: v for k, v in row.items() if k in field_set}
            metadata = KeyboardMetadata(**row_filtered)
        else:
            msg = "Unexpected"
            raise RuntimeError(msg)

    return Keyboard(meta=metadata, keys=keys)


def parse_ergogen_points(layout: dict) -> Keyboard:
    if not isinstance(layout, dict):
        msg = "Expected an object with nested objects"
        raise RuntimeError(msg)
    if not layout:
        msg = "Expected non-empty object"
        raise RuntimeError(msg)

    metadata: KeyboardMetadata = KeyboardMetadata()
    keys = []

    # in ergogen terminology, 'padding' is vertical space between each key
    # and 'spread' is horizontal space between each key. These values are defined
    # in 'meta' object for each 'point' (key). Let's assume that these values are
    # the same for all of them (which I don't know if it is actually true)
    padding = None
    spread = None

    # at the end of parsing will need to fix key order, kle starts from top,
    # ergogen from bottom
    # store topmost, leftmost key index while iterating, it will be used later
    topmost_leftmost = [0, 0]

    def __int(value: float) -> Union[int, float]:
        return int(value) if int(value) == value else value

    for item in layout.values():
        if "meta" not in item:
            msg = "Item needs to have meta defined"
            raise RuntimeError(msg)

        meta = item["meta"]

        # read key spacing only once, assume that it does not change
        if not padding and not spread:
            if "padding" not in meta or "spread" not in meta:
                msg = "Unable to determine key spacing"
                raise RuntimeError(msg)
            padding = meta["padding"]
            spread = meta["spread"]

        key = Key()

        # kle expresses distances (positions) with 1U, need to normalize
        # ergogen uses key center for position, kle uses key left top corner, need to adjust position by size
        key.x = __int(item["x"] / spread)
        key.y = __int(item["y"] / padding)
        key.width = __int(meta["width"] / spread)
        key.height = __int(meta["height"] / spread)

        # non-custom key shapes (like ISO enter) not supported:
        key.width2 = key.width
        key.height2 = key.height

        if key.y > topmost_leftmost[1] or (
            key.y == topmost_leftmost[1] and key.x <= topmost_leftmost[0]
        ):
            topmost_leftmost = [key.x, key.y]

        # kle and ergogen rotate in opposite directions
        key.rotation_angle = -1 * item["r"]

        keys.append(key)

    # do some cleanup to be kle compatible
    # adjust coordinates
    for key in keys:
        # reverse top-bottom
        key.y = abs(key.y - topmost_leftmost[1])
        # move position from key center (ergogen) to top left corner (kle)
        key.x = key.x - key.width / 2
        key.y = key.y - key.height / 2
    # move out of negative positions
    min_x = min(keys, key=lambda k: k.x).x
    if min_x >= 0:
        min_x = 0

    min_y = min(keys, key=lambda k: k.y).y
    if min_y >= 0:
        min_y = 0

    for key in keys:
        key.x = key.x - min_x
        key.y = key.y - min_y
        # rotation is always expressed in relation to key center
        key.rotation_x = key.x + key.width / 2 if key.rotation_angle else 0
        key.rotation_y = key.y + key.height / 2 if key.rotation_angle else 0

    # and sort (topmost leftmost first)
    keys = sorted(keys, key=lambda k: [k.y, k.x])

    return Keyboard(meta=metadata, keys=keys)


def get_keyboard(layout: dict) -> Keyboard:
    try:
        return parse_kle(layout)
    except Exception:
        pass
    try:
        return parse_via(layout)
    except Exception:
        pass
    try:
        return Keyboard.from_json(layout)
    except Exception:
        pass
    try:
        return parse_ergogen_points(layout)
    except Exception:
        pass
    msg = "Unable to get keyboard layout"
    raise RuntimeError(msg)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="KLE format converter")
    parser.add_argument("-in", required=True, help="Layout file")
    parser.add_argument(
        "-inform",
        required=False,
        default="RAW",
        choices=["KLE_RAW", "KLE_VIA", "KLE_INTERNAL", "ERGOGEN_INTERNAL"],
        help="Specifies the input format",
    )
    parser.add_argument("-out", required=False, help="Result file")
    parser.add_argument(
        "-outform",
        required=False,
        default="KLE_INTERNAL",
        choices=["KLE_RAW", "KLE_INTERNAL"],
        help="Specifies the output format",
    )
    parser.add_argument(
        "-text", required=False, action="store_true", help="Print result"
    )

    args = parser.parse_args()
    input_path = getattr(args, "in")
    input_format = args.inform
    output_path = getattr(args, "out")
    output_format = args.outform
    print_result = args.text

    if input_format == output_format:
        print("Output format equal input format, nothing to do...")
        sys.exit(1)

    with open(input_path, "r") as f:
        text_input = f.read()
        layout = json.loads(text_input)

        result = ""
        if input_format == "KLE_RAW":  # convert to KLE_INTERNAL
            result = parse_kle(layout)
            result = json.loads(result.to_json())
            if print_result:
                pprint.pprint(result)
        elif input_format == "KLE_VIA":  # convert to KLE_INTERNAL
            result = parse_via(layout)
            result = json.loads(result.to_json())
            if print_result:
                pprint.pprint(result)
        elif input_format == "KLE_INTERNAL":  # convert to KLE_RAW
            keyboard = Keyboard.from_json(layout)
            result = keyboard.to_kle()
            print(result)
            # 'to_kle' returns 'raw data' string which can be copy pasted
            # to keyboard-layout-editor, to make json out of it we need
            # to wrap it in list. Then it can be uploaded as JSON.
            result = "[" + result + "]"
            result = json.loads(result)
        else:  # ERGOGEN convert to KLE_RAW or KLE_INTERNAL
            result = parse_ergogen_points(layout)
            if output_format == "KLE_INTERNAL":
                result = json.loads(result.to_json())
                if print_result:
                    pprint.pprint(result)
            else:
                result = result.to_kle()
                if print_result:
                    print(result)
                result = "[" + result + "]"
                result = json.loads(result)

        if output_path:
            with open(output_path, "w") as f:
                json.dump(result, f, indent=4)



import argparse
import json
import logging
import os
import sys
from typing import List

import pcbnew

from .board_builder import BoardBuilder
from .defaults import DEFAULT_DIODE_POSITION, ZERO_POSITION
from .element_position import ElementInfo, ElementPosition, Point, PositionOption, Side
from .key_placer import KeyPlacer
from .kle_serial import parse_via
from .template_copier import TemplateCopier


class ElementInfoAction(argparse.Action):
    def __call__(self, parser, namespace, values: str, option_string=None) -> None:
        try:
            value: ElementInfo = self.parse(values, option_string)
        except ValueError as e:
            raise argparse.ArgumentTypeError(str(e))
        setattr(namespace, self.dest, value)

    def parse(self, values: str, option_string) -> ElementInfo:
        tokens: list[str] = values.split()
        err = ""
        if len(tokens) not in [2, 3, 6]:
            err = f"{option_string} invalid format."
            raise ValueError(err)
        else:
            annotation = tokens[0]
            if annotation.count("{}") != 1:
                err = (
                    f"'{annotation}' invalid annotation specifier, "
                    "it must contain exactly one '{}' placeholder."
                )
                raise ValueError(err)

            option = PositionOption.get(tokens[1])
            position = None
            template_path = ""

            if len(tokens) == 2:
                if option not in [PositionOption.RELATIVE, PositionOption.DEFAULT]:
                    err = (
                        f"{option_string} position option needs to be equal "
                        "RELATIVE or DEFAULT if position details not provided"
                    )
                    raise ValueError(err)
            elif len(tokens) == 3:
                if option not in [PositionOption.PRESET, PositionOption.RELATIVE]:
                    err = (
                        f"{option_string} position option needs to be equal"
                        "RELATIVE or PRESET when providing template path"
                    )
                    raise ValueError(err)
            elif option != PositionOption.CUSTOM:
                err = (
                    f"{option_string} position option needs to be equal CUSTOM "
                    "when providing position details"
                )
                raise ValueError(err)

            if option == PositionOption.CUSTOM:
                floats = tuple(map(float, tokens[2:5]))
                side = Side.get(tokens[5])
                position = ElementPosition(Point(floats[0], floats[1]), floats[2], side)
            elif option == PositionOption.RELATIVE:
                # template path if optional for RELATIVE option:
                position = None
                if len(tokens) == 3:
                    template_path = tokens[2]
            elif option == PositionOption.PRESET:
                position = None
                template_path = tokens[2]

            value: ElementInfo = ElementInfo(
                annotation,
                option,
                position,
                template_path,
            )
            return value


class ElementInfoListAction(ElementInfoAction):
    def __call__(self, parser, namespace, values: str, option_string=None) -> None:
        try:
            if "DEFAULT" in values:
                msg = f"{option_string} does not support DEFAULT position"
                raise ValueError(msg)
            value: List[ElementInfo] = []
            tokens: list[str] = values.split(";")
            for token in tokens:
                element_info: ElementInfo = super().parse(token.strip(), option_string)
                value.append(element_info)
        except ValueError as e:
            raise argparse.ArgumentTypeError(str(e))
        setattr(namespace, self.dest, value)


class XYAction(argparse.Action):
    def __call__(self, parser, namespace, values: str, option_string=None) -> None:
        try:
            value = tuple(map(float, values.split()))
            if len(value) != 2:
                msg = (
                    f"{option_string} must be exactly two numeric values "
                    "separated by a space."
                )
                raise ValueError(msg)
        except ValueError as e:
            raise argparse.ArgumentTypeError(str(e))
        setattr(namespace, self.dest, value)


def app():
    parser = argparse.ArgumentParser(
        description="Keyboard's key autoplacer",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "-b",
        "--board",
        required=True,
        help=".kicad_pcb file to be processed or created",
    )
    parser.add_argument(
        "-l", "--layout", required=False, help="json layout definition file"
    )
    parser.add_argument(
        "--route-switches-with-diodes",
        action="store_true",
        help="Enable switch-diode routing",
    )
    parser.add_argument(
        "--route-rows-and-columns",
        action="store_true",
        help="Enable rows/columns routing",
    )
    parser.add_argument(
        "-d",
        "--diode",
        default=ElementInfo("D{}", PositionOption.DEFAULT, DEFAULT_DIODE_POSITION, ""),
        action=ElementInfoAction,
        help=(
            "Diode information, space separated value of ANNOTATION OPTION [POSITION]\n"
            "Available OPTION choices: DEFAULT, RELATIVE, PRESET and CUSTOM\n"
            "When DEFAULT, then POSITION needs to be omitted,\n"
            "when RELATIVE, then POSITION is optional path for saving kicad_pcb template file\n"
            "when PRESET, then POSITION is mandatory path to kicad_pcb template file\n"
            "when CUSTOM, then POSITION is space separated value of X Y ORIENTATION FRONT|BACK\n"
            "for example:\n"
            "\tD{} RELATIVE\n"
            "\tD{} PRESET /home/user/project/diode_preset.kicad_pcb\n"
            "\tD{} CUSTOM 5 -4.5 90 BACK\n"
            "equal 'D{} DEFAULT' by default"
        ),
    )
    parser.add_argument(
        "--additional-elements",
        default=[
            ElementInfo(
                "ST{}",
                PositionOption.CUSTOM,
                ZERO_POSITION,
                "",
            )
        ],
        action=ElementInfoListAction,
        help=(
            "List of ';' separated additional elements ELEMENT_INFO values\n"
            "ELEMENT_INFO is space separated value of ANNOTATION OPTION POSITION\n"
            "Available OPTION choices: RELATIVE, PRESET and CUSTOM\n"
            "when RELATIVE, then POSITION is optional path for saving kicad_pcb template file\n"
            "when PRESET, then POSITION is mandatory path to kicad_pcb template file\n"
            "when CUSTOM, then POSITION is space separated value of X Y ORIENTATION FRONT|BACK\n"
            "for example:\n"
            "\tST{} CUSTOM 0 0 180 BACK;LED{} RELATIVE\n"
            "equal 'ST{} CUSTOM 0 0 0 FRONT' by default"
        ),
    )
    parser.add_argument(
        "--key-distance",
        default=(19.05, 19.05),
        action=XYAction,
        help=(
            "X and Y key 1U distance in mm, as two space separated numeric values, "
            "19.05 19.05 by default"
        ),
    )
    parser.add_argument("-t", "--template", help="Controller circuit template")
    parser.add_argument(
        "--create-from-via",
        required=False,
        action="store_true",
        help=(
            "Enables experimental creation mode, where via-annotated kle layout is used\n"
            "for adding footprints and netlists to the newly created board file.\n"
        ),
    )
    parser.add_argument(
        "--switch-footprint",
        required=False,
        type=str,
        help=(
            "Full path to switch footprint, required when `--crate-from-via` option used."
        ),
    )
    parser.add_argument(
        "--diode-footprint",
        required=False,
        type=str,
        help=(
            "Full path to diode footprint, required when `--crate-from-via` option used."
        ),
    )

    args = parser.parse_args()

    layout_path = args.layout
    board_path = args.board
    route_switches_with_diodes = args.route_switches_with_diodes
    route_rows_and_columns = args.route_rows_and_columns
    diode = args.diode
    additional_elements = args.additional_elements
    key_distance = args.key_distance
    template_path = args.template
    create_from_via = args.create_from_via

    # set up logger
    logging.basicConfig(
        level=logging.DEBUG, format="%(asctime)s: %(message)s", datefmt="%H:%M:%S"
    )

    if layout_path:
        with open(layout_path, "r") as f:
            layout = json.load(f)
    else:
        layout = {}

    if create_from_via:
        if os.path.isfile(board_path):
            logging.error(f"File {board_path} already exist, aborting")
            sys.exit(1)

        builder = BoardBuilder(
            switch_footprint=args.switch_footprint,
            diode_footprint=args.diode_footprint,
        )
        board = builder.create_board(parse_via(layout))
        board.Save(board_path)

    board = pcbnew.LoadBoard(board_path)

    placer = KeyPlacer(board, key_distance)
    placer.run(
        layout,
        "SW{}",
        diode,
        route_switches_with_diodes,
        route_rows_and_columns,
        additional_elements=additional_elements,
    )

    if template_path:
        copier = TemplateCopier(board, template_path, route_rows_and_columns)
        copier.run()

    pcbnew.Refresh()
    pcbnew.SaveBoard(board_path, board)

    logging.shutdown()


if __name__ == "__main__":
    app()



from __future__ import annotations

import gettext
import json
import logging
import os
import string
import sys
from enum import Flag
from typing import List, Optional, Tuple

import wx
from wx.lib.embeddedimage import PyEmbeddedImage

from .defaults import DEFAULT_DIODE_POSITION, ZERO_POSITION
from .element_position import ElementInfo, ElementPosition, Point, PositionOption, Side
from .help_dialog import HelpDialog

logger = logging.getLogger(__name__)
TEXT_CTRL_EXTRA_SPACE = 25

# Most of the phrases used in this plugin are already in use in KiCad.
# It means that we get translations for free using `wx.GetTranslation`.
# All strings translated with 'wx_' are expected to be a part of
# KiCad's translation files. All remaining will be translated with
# another custom mechanism or will remain default.
wx_ = wx.GetTranslation

# Currently there is no elegant way to check which language is loaded by KiCad.
# This feature has been requested here:
#   https://gitlab.com/kicad/code/kicad/-/issues/10573
# Until then, use workaroud - request translation with wx_ and use result
# in lookup table. This lookup should contain all installed languages defined
# in translation/pofiles/LINGUAS_INSTALL.
KICAD_TRANSLATIONS_LOOKUP = {
    "Sprache": "de",
    "Set Language": "en",
    "Seleccionar idioma": "es",
    "言語設定": "ja",
    "언어 설정": "ko",
    "Ustaw język": "pl",
    "Установить язык": "ru",
    "設定語言": "zh_CN",
}


def get_current_kicad_language():
    return KICAD_TRANSLATIONS_LOOKUP.get(wx_("Set Language"), "en")


def get_plugin_translator(lang: str = "en"):
    localedir = os.path.join(os.path.dirname(__file__), "locale")
    trans = gettext.translation(
        "kbplacer", localedir=localedir, languages=(lang,), fallback=True
    )
    return trans.gettext


def get_file_picker(*args, **kwargs):
    file_picker = wx.FilePickerCtrl(*args, **kwargs)
    file_picker.SetTextCtrlGrowable(True)

    def _update_position(_):
        text_ctrl = file_picker.GetTextCtrl()
        # when updating from file chooser window we want automatically
        # move to the end (it looks better when not whole path fits),
        # but when user updates by typing in text control we don't
        # want to mess up with it. This can be done by checking
        # if current insertion point is 0.
        if text_ctrl.GetInsertionPoint() == 0:
            text_ctrl.SetInsertionPointEnd()

    file_picker.Bind(wx.EVT_FILEPICKER_CHANGED, _update_position)
    return file_picker


class FloatValidator(wx.Validator):
    def __init__(self) -> None:
        wx.Validator.__init__(self)
        self.Bind(wx.EVT_CHAR, self.OnChar)

    def Clone(self) -> FloatValidator:
        return FloatValidator()

    def Validate(self, _):
        text_ctrl = self.GetWindow()
        if not text_ctrl.IsEnabled():
            return True

        text = text_ctrl.GetValue()
        try:
            float(text)
            return True
        except ValueError:
            # this can happen when value is empty, equal '-', '.', or '-.',
            # other invalid values should not be allowed by 'OnChar' filtering
            name = text_ctrl.GetName()
            wx.MessageBox(f"Invalid '{name}' float value: '{text}'!", "Error")
            text_ctrl.SetFocus()
            return False

    def TransferToWindow(self):
        return True

    def TransferFromWindow(self):
        return True

    def OnChar(self, event):
        text_ctrl = self.GetWindow()
        current_position = text_ctrl.GetInsertionPoint()
        keycode = int(event.GetKeyCode())
        if keycode in [
            wx.WXK_BACK,
            wx.WXK_DELETE,
            wx.WXK_LEFT,
            wx.WXK_RIGHT,
            wx.WXK_NUMPAD_LEFT,
            wx.WXK_NUMPAD_RIGHT,
        ]:
            event.Skip()
        else:
            text_ctrl = self.GetWindow()
            text = text_ctrl.GetValue()
            key = chr(keycode)
            if (
                # allow only digits
                # or single '-' when as first character
                # or single '.'
                key in string.digits
                or (key == "-" and "-" not in text and current_position == 0)
                or (key == "." and "." not in text)
            ):
                event.Skip()


class LabeledTextCtrl(wx.Panel):
    def __init__(
        self,
        parent,
        label: str,
        value: str,
        width: int = -1,
        validator: wx.Validator = wx.DefaultValidator,
    ) -> None:
        super().__init__(parent)

        expected_char_width = self.GetTextExtent("x").x
        if width != -1:
            annotation_format_size = wx.Size(
                expected_char_width * width + TEXT_CTRL_EXTRA_SPACE, -1
            )
        else:
            annotation_format_size = wx.Size(-1, -1)

        self.label = wx.StaticText(self, -1, label)
        self.text = wx.TextCtrl(
            self,
            value=value,
            size=annotation_format_size,
            validator=validator,
            name=label.strip(":"),
        )

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self.label, 0, wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 5)
        sizer.Add(self.text, 0, wx.EXPAND | wx.TOP | wx.BOTTOM, 5)

        self.SetSizer(sizer)

    def Enable(self):
        self.label.Enable()
        self.text.Enable()

    def Disable(self):
        self.label.Disable()
        self.text.Disable()


class CustomRadioBox(wx.Panel):
    def __init__(self, parent, choices: List[str]) -> None:
        super().__init__(parent)
        self.radio_buttons: dict[str, wx.RadioButton] = {}

        for choice in choices:
            radio_button = wx.RadioButton(self, label=choice)
            self.radio_buttons[choice] = radio_button

        # this is special hidden option to allow clearing (selecting none)
        self.none_button = wx.RadioButton(self, label="")
        self.none_button.Hide()

        # on linux, use negative border because otherwise it looks too spaced out
        space = 0 if sys.platform == "win32" else -2
        sizer = wx.BoxSizer(wx.VERTICAL)
        for radio_button in self.radio_buttons.values():
            sizer.Add(radio_button, 0, wx.TOP | wx.BOTTOM, space)

        self.SetSizer(sizer)

    def Select(self, choice):
        self.radio_buttons[choice].SetValue(True)

    def Clear(self):
        self.none_button.SetValue(True)

    def GetValue(self) -> Optional[str]:
        if not self.none_button.GetValue():
            for choice, button in self.radio_buttons.items():
                if button.GetValue():
                    return choice
        return None


class ElementPositionWidget(wx.Panel):
    def __init__(
        self,
        parent,
        default_position: Optional[ElementPosition] = None,
    ) -> None:
        super().__init__(parent)

        self.default = default_position
        self.x = LabeledTextCtrl(
            self, wx_("Offset X:"), value="", width=5, validator=FloatValidator()
        )
        self.y = LabeledTextCtrl(
            self, wx_("Y:"), value="", width=5, validator=FloatValidator()
        )
        self.orientation = LabeledTextCtrl(
            self, wx_("Orientation:"), value="", width=5, validator=FloatValidator()
        )
        self.side_label = wx.StaticText(self, -1, wx_("Side:"))
        self.side = CustomRadioBox(self, choices=[wx_("Front"), wx_("Back")])

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self.x, 0, wx.EXPAND | wx.LEFT, 5)
        sizer.Add(self.y, 0, wx.EXPAND | wx.LEFT, 5)
        sizer.Add(self.orientation, 0, wx.EXPAND | wx.LEFT, 5)
        sizer.Add(self.side_label, 0, wx.LEFT | wx.ALIGN_CENTER_VERTICAL, 5)
        sizer.Add(self.side, 0, wx.EXPAND | wx.LEFT, 5)

        self.SetSizer(sizer)

    def set_position_by_choice(self, choice: str) -> None:
        if choice == PositionOption.DEFAULT:
            self.__set_position_to_default()
        elif choice in [PositionOption.RELATIVE, PositionOption.PRESET]:
            self.__set_position_to_empty_non_editable()
        elif choice == PositionOption.CUSTOM:
            self.__set_position_to_zero_editable()
        else:
            raise ValueError

    def set_position(self, position: ElementPosition) -> None:
        self.__set_coordinates(
            str(position.relative_position.x), str(position.relative_position.y)
        )
        self.__set_orientation(str(position.orientation))
        self.__set_side(position.side)

    def __set_position_to_default(self) -> None:
        if self.default:
            x = str(self.default.relative_position.x)
            y = str(self.default.relative_position.y)
            self.__set_coordinates(x, y)
            self.__set_orientation(str(self.default.orientation))
            self.__set_side(self.default.side)
            self.Disable()

    def __set_position_to_zero_editable(self) -> None:
        self.__set_coordinates("0", "0")
        self.__set_orientation("0")
        self.__set_side(Side.FRONT)
        self.Enable()

    def __set_position_to_empty_non_editable(self) -> None:
        self.__set_coordinates("-", "-")
        self.__set_orientation("-")
        self.side.Clear()
        self.Disable()

    def __set_coordinates(self, x: str, y: str) -> None:
        self.x.text.SetValue(x)
        self.y.text.SetValue(y)

    def __set_orientation(self, orientation: str) -> None:
        self.orientation.text.SetValue(orientation)

    def __set_side(self, side: Side) -> None:
        if side == Side.BACK:
            self.side.Select(wx_("Back"))
        else:
            self.side.Select(wx_("Front"))

    def GetValue(self) -> ElementPosition:
        x = float(self.x.text.GetValue())
        y = float(self.y.text.GetValue())
        orientation = float(self.orientation.text.GetValue())
        side_str = self.side.GetValue()
        return ElementPosition(Point(x, y), orientation, Side(side_str == wx_("Back")))

    def Enable(self):
        self.x.Enable()
        self.y.Enable()
        self.orientation.Enable()
        self.side_label.Enable()
        self.side.Enable()

    def Disable(self):
        self.x.Disable()
        self.y.Disable()
        self.orientation.Disable()
        self.side_label.Disable()
        self.side.Disable()


class TemplateType(Flag):
    LOAD = False
    SAVE = True


class ElementTemplateSelectionWidget(wx.Panel):
    def __init__(
        self, parent, picker_type: TemplateType, initial_path: str = ""
    ) -> None:
        super().__init__(parent)
        self._ = self.GetTopLevelParent()._

        label = (
            self._("Save to:")
            if picker_type == TemplateType.SAVE
            else self._("Load from:")
        )
        layout_label = wx.StaticText(self, -1, label)
        layout_picker = get_file_picker(
            self,
            -1,
            wildcard="KiCad printed circuit board files (*.kicad_pcb)|*.kicad_pcb",
            style=wx.FLP_USE_TEXTCTRL
            | (wx.FLP_SAVE if picker_type == TemplateType.SAVE else wx.FLP_OPEN),
        )

        if initial_path:
            layout_picker.SetPath(initial_path)
            layout_picker.GetTextCtrl().SetInsertionPointEnd()

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(layout_label, 0, wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 5)
        sizer.Add(layout_picker, 1, wx.EXPAND | wx.ALL, 5)
        self.SetSizer(sizer)

        self.__layout_picker = layout_picker
        self.__picker_type = picker_type

    def GetValue(self) -> str:
        path = self.__layout_picker.GetPath()
        if self.__picker_type == TemplateType.LOAD and not path:
            # can't be empty when running in `LOAD` mode
            msg = "'Load from:' file picker can't be empty"
            raise ValueError(msg)
        return path


class ElementPositionChoiceWidget(wx.Panel):
    def __init__(
        self,
        parent,
        initial_choice: PositionOption,
        initial_position: Optional[ElementPosition] = None,
        default_position: Optional[ElementPosition] = None,
        initial_path: str = "",
    ) -> None:
        super().__init__(parent)

        choices = [
            PositionOption.CUSTOM,
            PositionOption.RELATIVE,
            PositionOption.PRESET,
        ]
        if default_position:
            choices.insert(0, PositionOption.DEFAULT)

        self.dropdown = wx.ComboBox(self, choices=choices, style=wx.CB_DROPDOWN)
        self.dropdown.Bind(wx.EVT_COMBOBOX, self.__on_position_choice_change)

        dropdown_sizer = wx.BoxSizer(wx.HORIZONTAL)
        dropdown_sizer.Add(
            wx.StaticText(self, -1, wx_("Position") + ":"),
            0,
            wx.RIGHT | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        dropdown_sizer.Add(self.dropdown, 0, wx.EXPAND | wx.TOP | wx.BOTTOM, 5)

        self.position = ElementPositionWidget(self, default_position)
        self.load_template = ElementTemplateSelectionWidget(
            self, picker_type=TemplateType.LOAD, initial_path=initial_path
        )
        self.save_template = ElementTemplateSelectionWidget(
            self, picker_type=TemplateType.SAVE, initial_path=initial_path
        )

        self.__set_initial_state(initial_choice)
        if initial_position and initial_choice == PositionOption.CUSTOM:
            self.position.set_position(initial_position)

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(dropdown_sizer, 0, wx.EXPAND | wx.ALL, 5)
        sizer.Add(self.position, 0, wx.EXPAND | wx.ALL, 5)
        sizer.Add(self.load_template, 1, wx.EXPAND | wx.ALL, 5)
        sizer.Add(self.save_template, 1, wx.EXPAND | wx.ALL, 5)

        self.SetSizer(sizer)

    def __set_initial_state(self, choice) -> None:
        self.dropdown.SetValue(choice)
        self.__set_position_by_choice(choice)

    def __on_position_choice_change(self, event) -> None:
        choice = event.GetString()
        self.__set_position_by_choice(choice)

    def __set_position_by_choice(self, choice: str) -> None:
        if choice == PositionOption.RELATIVE:
            self.position.Hide()
            self.load_template.Hide()
            self.save_template.Show()
        elif choice == PositionOption.PRESET:
            self.position.Hide()
            self.load_template.Show()
            self.save_template.Hide()
        else:
            self.position.Show()
            self.load_template.Hide()
            self.save_template.Hide()
        self.GetTopLevelParent().Layout()
        self.position.set_position_by_choice(choice)
        self.choice = PositionOption(choice)

    def GetValue(self) -> Tuple[PositionOption, Optional[ElementPosition], str]:
        template_path = ""
        if self.dropdown.GetValue() == PositionOption.RELATIVE:
            template_path = self.save_template.GetValue()
        elif self.dropdown.GetValue() == PositionOption.PRESET:
            template_path = self.load_template.GetValue()

        if self.choice not in [PositionOption.DEFAULT, PositionOption.CUSTOM]:
            return self.choice, None, template_path
        return self.choice, self.position.GetValue(), template_path

    def Enable(self):
        self.dropdown.Enable()
        if self.dropdown.GetValue() == PositionOption.CUSTOM:
            self.position.Enable()

    def Disable(self):
        self.dropdown.Disable()
        self.position.Disable()


class ElementSettingsWidget(wx.Panel):
    def __init__(
        self,
        parent,
        element_info: ElementInfo,
        default_position: Optional[ElementPosition] = None,
    ) -> None:
        super().__init__(parent)

        self.annotation_format = LabeledTextCtrl(
            self,
            label=wx_("Footprint Annotation") + ":",
            value=element_info.annotation_format,
            width=3,
        )
        self.position_widget = ElementPositionChoiceWidget(
            self,
            element_info.position_option,
            element_info.position,
            default_position=default_position,
            initial_path=element_info.template_path,
        )

        sizer = wx.BoxSizer(wx.HORIZONTAL)

        sizer.Add(
            self.annotation_format, 0, wx.EXPAND | wx.TOP | wx.BOTTOM | wx.RIGHT, 5
        )
        sizer.Add(self.position_widget, 1, wx.EXPAND | wx.ALL, 0)

        self.SetSizer(sizer)

    def GetValue(self) -> ElementInfo:
        annotation = self.annotation_format.text.GetValue()
        position = self.position_widget.GetValue()
        return ElementInfo(annotation, *position)

    def Enable(self):
        self.position_widget.Enable()

    def Disable(self):
        self.position_widget.Disable()


class KbplacerDialog(wx.Dialog):
    def __init__(self, parent, title, initial_state: Optional[dict] = None) -> None:
        style = wx.DEFAULT_DIALOG_STYLE
        super(KbplacerDialog, self).__init__(parent, -1, title, style=style)

        language = get_current_kicad_language()
        logger.info(f"Language: {language}")
        self._ = get_plugin_translator(language)

        def __get_params(name) -> dict:
            return initial_state.get(name, {}) if initial_state else {}

        params: dict = __get_params("switch_section")
        switch_section: wx.Sizer = self.get_switch_section(**params)

        params: dict = __get_params("switch_diodes_section")
        if "element_info" in params:
            try:
                params["element_info"] = ElementInfo.from_dict(params["element_info"])
            except:
                params = {}
        switch_diodes_section: wx.Sizer = self.get_switch_diodes_section(**params)

        params: dict = __get_params("additional_elements")
        if "elements_info" in params:
            try:
                params["elements_info"] = [
                    ElementInfo.from_dict(v) for v in params["elements_info"]
                ]
            except:
                params = {}
        additional_elements_section: wx.Sizer = self.get_additional_elements_section(
            **params
        )

        params: dict = __get_params("misc_section")
        misc_section: wx.Sizer = self.get_misc_section(**params)

        box: wx.Sizer = wx.BoxSizer(wx.VERTICAL)

        box.Add(switch_section, 0, wx.EXPAND | wx.ALL, 5)
        box.Add(switch_diodes_section, 0, wx.EXPAND | wx.ALL, 5)
        box.Add(additional_elements_section, 0, wx.EXPAND | wx.ALL, 5)
        box.Add(misc_section, 0, wx.EXPAND | wx.ALL, 5)

        buttons = self.CreateButtonSizer(wx.OK | wx.CANCEL | wx.HELP)

        if help_button := wx.FindWindowById(wx.ID_HELP, self):
            help_button.Bind(wx.EVT_BUTTON, self.on_help_button)

        box.Add(buttons, 0, wx.EXPAND | wx.ALL, 5)

        self.SetSizerAndFit(box)

    def get_switch_section(
        self,
        annotation: str = "SW{}",
        layout_path: str = "",
        x_distance: str = "19.05",
        y_distance: str = "19.05",
    ) -> wx.Sizer:
        key_annotation = LabeledTextCtrl(
            self, wx_("Footprint Annotation") + ":", annotation
        )

        layout_label = wx.StaticText(self, -1, self._("Keyboard layout file:"))
        layout_picker = get_file_picker(
            self,
            -1,
            wildcard="JSON files (*.json)|*.json|All files (*)|*",
            style=wx.FLP_USE_TEXTCTRL,
        )
        if layout_path:
            layout_picker.SetPath(layout_path)
            layout_picker.GetTextCtrl().SetInsertionPointEnd()

        key_distance_x = LabeledTextCtrl(
            self, wx_("Step X:"), value=x_distance, width=5, validator=FloatValidator()
        )
        key_distance_y = LabeledTextCtrl(
            self, wx_("Step Y:"), value=y_distance, width=5, validator=FloatValidator()
        )

        box = wx.StaticBox(self, label=self._("Switch settings"))
        sizer = wx.StaticBoxSizer(box, wx.HORIZONTAL)
        sizer.Add(key_annotation, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer.Add(layout_label, 0, wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 5)
        sizer.Add(layout_picker, 1, wx.ALL, 5)
        sizer.Add(key_distance_x, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer.Add(key_distance_y, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)

        self.__key_annotation_format = key_annotation.text
        self.__layout_picker = layout_picker
        self.__key_distance_x = key_distance_x.text
        self.__key_distance_y = key_distance_y.text

        return sizer

    def get_switch_diodes_section(
        self,
        enable: bool = True,
        route_switches_with_diodes: bool = True,
        element_info: ElementInfo = ElementInfo(
            "D{}", PositionOption.DEFAULT, DEFAULT_DIODE_POSITION, ""
        ),
    ) -> wx.Sizer:
        place_diodes_checkbox = wx.CheckBox(self, label=wx_("Allow autoplacement"))
        place_diodes_checkbox.SetValue(enable)
        place_diodes_checkbox.Bind(wx.EVT_CHECKBOX, self.on_diode_place_checkbox)

        switches_and_diodes_tracks_checkbox = wx.CheckBox(
            self, label=self._("Route with switches")
        )
        switches_and_diodes_tracks_checkbox.SetValue(route_switches_with_diodes)

        diode_settings = ElementSettingsWidget(
            self,
            element_info,
            default_position=DEFAULT_DIODE_POSITION,
        )

        box = wx.StaticBox(self, label=self._("Switch diodes settings"))
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)

        row1 = wx.BoxSizer(wx.HORIZONTAL)
        row1.Add(place_diodes_checkbox, 0, wx.ALL, 0)
        row1.Add(switches_and_diodes_tracks_checkbox, 0, wx.ALL, 0)

        sizer.Add(row1, 0, wx.EXPAND | wx.ALL, 5)
        # weird border value to make it aligned with 'additional_elements_section':
        sizer.Add(diode_settings, 0, wx.EXPAND | wx.ALL, 9)

        self.__place_diodes_checkbox = place_diodes_checkbox
        self.__switches_and_diodes_tracks_checkbox = switches_and_diodes_tracks_checkbox
        self.__diode_settings = diode_settings

        self.__enable_diode_settings(enable)

        return sizer

    def __enable_diode_settings(self, enable):
        if enable:
            self.__diode_settings.Enable()
        else:
            self.__diode_settings.Disable()

    def on_diode_place_checkbox(self, event):
        is_checked = event.GetEventObject().IsChecked()
        self.__enable_diode_settings(is_checked)

    def get_additional_elements_section(
        self,
        elements_info: List[ElementInfo] = [
            ElementInfo("ST{}", PositionOption.CUSTOM, ZERO_POSITION, "")
        ],
    ) -> wx.Sizer:
        self.__additional_elements = []

        scrolled_window = wx.ScrolledWindow(self)
        scrolled_window_sizer = wx.BoxSizer(wx.VERTICAL)

        scrolled_window.SetSizer(scrolled_window_sizer)
        virtual_width, virtual_height = scrolled_window_sizer.GetMinSize()
        scrolled_window.SetVirtualSize((virtual_width, virtual_height))
        scrolled_window.SetScrollRate(0, 10)

        box = wx.StaticBox(self, label=self._("Additional elements settings"))
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)

        sizer.Add(scrolled_window, 1, wx.EXPAND | wx.ALL, 10)

        dialog_width, _ = self.GetSize()
        sizer.SetMinSize((dialog_width, 180))

        buttons_sizer = wx.BoxSizer(wx.HORIZONTAL)

        def add_element(element_info: ElementInfo) -> None:
            element_settings = ElementSettingsWidget(scrolled_window, element_info)
            self.__additional_elements.append(element_settings)
            scrolled_window_sizer.Add(element_settings, 0, wx.EXPAND | wx.ALIGN_LEFT, 0)
            self.GetTopLevelParent().Layout()

        def add_element_callback(_) -> None:
            add_element(ElementInfo("", PositionOption.CUSTOM, ZERO_POSITION, ""))

        for element_info in elements_info:
            add_element(element_info)

        add_icon = PyEmbeddedImage(
            b"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAACtJ"
            b"REFUOI1jYKAx+A/FOAETpTaMGsDAwAil8YY0Pv0Uu4AQGE0H9DCAYgAADfAFFDV6vY8AAAAA"
            b"SUVORK5CYII="
        ).GetBitmap()
        add_button = wx.BitmapButton(self, bitmap=add_icon)
        add_button.Bind(wx.EVT_BUTTON, add_element_callback)

        def remove_element(_) -> None:
            element_settings = (
                self.__additional_elements.pop() if self.__additional_elements else None
            )
            if element_settings:
                element_settings.Destroy()
                self.Layout()

        remove_icon = PyEmbeddedImage(
            b"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABtJ"
            b"REFUOI1jYBgFwwAwQun/5OpnopZLRsGQBgBLTwEEpzJYVwAAAABJRU5ErkJggg=="
        ).GetBitmap()
        remove_button = wx.BitmapButton(self, bitmap=remove_icon)
        remove_button.Bind(wx.EVT_BUTTON, remove_element)

        buttons_sizer.Add(add_button, 0, wx.EXPAND | wx.ALL, 0)
        buttons_sizer.Add(remove_button, 0, wx.EXPAND | wx.ALL, 0)

        sizer.Add(buttons_sizer, 0, wx.EXPAND | wx.ALL, 5)

        return sizer

    def get_misc_section(
        self, route_rows_and_columns: bool = True, template_path: str = ""
    ) -> wx.Sizer:
        row_and_columns_tracks_checkbox = wx.CheckBox(
            self, label=self._("Route rows and columns")
        )
        row_and_columns_tracks_checkbox.SetValue(route_rows_and_columns)

        template_label = wx.StaticText(
            self, -1, self._("Controller circuit template file:")
        )
        template_picker = get_file_picker(
            self,
            -1,
            wildcard="KiCad printed circuit board files (*.kicad_pcb)|*.kicad_pcb",
            style=wx.FLP_USE_TEXTCTRL,
        )
        if template_path:
            template_picker.SetPath(template_path)
            template_picker.GetTextCtrl().SetInsertionPointEnd()

        box = wx.StaticBox(self, label=self._("Other settings"))
        sizer = wx.StaticBoxSizer(box, wx.HORIZONTAL)

        sizer.Add(row_and_columns_tracks_checkbox, 0, wx.EXPAND | wx.ALL, 5)
        sizer.Add(wx.StaticLine(self, style=wx.LI_VERTICAL), 0, wx.EXPAND | wx.ALL, 5)
        sizer.Add(template_label, 0, wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 5)
        sizer.Add(template_picker, 1, wx.EXPAND | wx.ALL, 5)

        self.__rows_and_columns_tracks_checkbox = row_and_columns_tracks_checkbox
        self.__template_picker = template_picker

        return sizer

    def on_help_button(self, event):
        del event
        help_dialog = HelpDialog(self)
        help_dialog.ShowModal()
        help_dialog.Destroy()

    def get_layout_path(self) -> str:
        return self.__layout_picker.GetPath()

    def get_key_annotation_format(self) -> str:
        return self.__key_annotation_format.GetValue()

    def route_switches_with_diodes(self) -> bool:
        return self.__switches_and_diodes_tracks_checkbox.GetValue()

    def route_rows_and_columns(self) -> bool:
        return self.__rows_and_columns_tracks_checkbox.GetValue()

    def get_key_distance(self) -> Tuple[float, float]:
        x = float(self.__key_distance_x.GetValue())
        y = float(self.__key_distance_y.GetValue())
        return x, y

    def get_template_path(self) -> str:
        return self.__template_picker.GetPath()

    def get_diode_position_info(
        self,
    ) -> Optional[ElementInfo]:
        return self.__diode_settings.GetValue()

    def get_additional_elements_info(
        self,
    ) -> List[ElementInfo]:
        return [
            e.GetValue()
            for e in self.__additional_elements
            if e.GetValue().annotation_format != ""
        ]

    def get_window_state(self):
        window_state = {
            "switch_section": {
                "annotation": self.get_key_annotation_format(),
                "layout_path": self.get_layout_path(),
                "x_distance": self.__key_distance_x.GetValue(),
                "y_distance": self.__key_distance_y.GetValue(),
            },
            "switch_diodes_section": {
                "enable": self.__place_diodes_checkbox.GetValue(),
                "route_switches_with_diodes": self.route_switches_with_diodes(),
                "element_info": self.__diode_settings.GetValue().to_dict(),
            },
            "additional_elements": {
                "elements_info": [
                    e.GetValue().to_dict() for e in self.__additional_elements
                ],
            },
            "misc_section": {
                "route_rows_and_columns": self.route_rows_and_columns(),
                "template_path": self.get_template_path(),
            },
        }
        return json.dumps(window_state, indent=None)


# used for tests
if __name__ == "__main__":
    import argparse
    import threading

    parser = argparse.ArgumentParser(description="dialog test")
    parser.add_argument("-i", "--initial-state", default="{}", help="Initial gui state")
    parser.add_argument(
        "-o", "--output-dir", required=True, help="Directory for output files"
    )
    args = parser.parse_args()

    initial_state = json.loads(args.initial_state)
    _ = wx.App()
    dlg = KbplacerDialog(None, "kbplacer", initial_state=initial_state)
    with open(f"{args.output_dir}/window_state.json", "w") as f:
        f.write(dlg.get_window_state())

    if "PYTEST_CURRENT_TEST" in os.environ:
        # use stdin for gracefully closing GUI when running
        # from pytest. This is required when measuring
        # coverage and process kill would cause measurement to be lost
        def listen_for_exit():
            while True:
                input("Press any key to exit: ")
                dlg.Close(True)
                sys.exit()

        input_thread = threading.Thread(target=listen_for_exit)
        input_thread.daemon = True
        input_thread.start()

    dlg.ShowModal()



import os
import re
import webbrowser

import wx

wx_ = wx.GetTranslation


class HelpDialog(wx.Dialog):
    def __init__(self, parent) -> None:
        super(HelpDialog, self).__init__(parent, -1, "kbplacer help")

        # inherit translations from parent:
        self._ = parent._ if parent else (lambda x: x)
        information_section = self.get_information_section()
        actions_section = self.get_actions_section()
        help_section = self.get_help_section()

        buttons = self.CreateButtonSizer(wx.OK)

        header = wx.BoxSizer(wx.HORIZONTAL)
        header.Add(information_section, 3, wx.ALL, 5)
        header.Add(actions_section, 1, wx.ALIGN_CENTER | wx.ALL, 5)

        box = wx.BoxSizer(wx.VERTICAL)
        box.Add(header, 0, wx.EXPAND | wx.ALL, 5)
        box.Add(help_section, 0, wx.EXPAND | wx.ALL, 5)
        box.Add(buttons, 0, wx.EXPAND | wx.ALL, 5)

        self.SetSizerAndFit(box)

    def get_information_section(self):
        source_dir = os.path.dirname(__file__)
        icon_file_name = os.path.join(source_dir, "icon.png")
        icon = wx.Image(icon_file_name, wx.BITMAP_TYPE_ANY)
        icon_bitmap = wx.Bitmap(icon)
        static_icon_bitmap = wx.StaticBitmap(self, wx.ID_ANY, icon_bitmap)

        font = wx.Font(
            12, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_BOLD
        )
        name = wx.StaticText(self, -1, "Keyboard Footprints Placer")
        name.SetFont(font)

        version_file_name = os.path.join(source_dir, "version.txt")
        version_str = self._("<missing>")
        if os.path.isfile(version_file_name):
            with open(version_file_name, "r") as f:
                version_str = f.read()
        if not re.match(r"v\d.\d$", version_str):
            status = ", " + self._("development build")
        else:
            status = ""
        version = wx.StaticText(self, -1, wx_("Version") + f": {version_str}{status}")

        name_box = wx.BoxSizer(wx.HORIZONTAL)
        name_box.Add(static_icon_bitmap, 0, wx.ALL, 5)
        name_box.Add(name, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)

        box = wx.BoxSizer(wx.VERTICAL)
        box.Add(name_box, 0, wx.ALL, 5)
        box.Add(version, 0, wx.ALL, 5)

        return box

    def get_actions_section(self):
        box = wx.BoxSizer(wx.VERTICAL)

        report_bug_button = wx.Button(self, label=wx_("Report Bug"))

        def on_report_bug(_) -> None:
            webbrowser.open(
                "https://github.com/adamws/kicad-kbplacer/issues/new?template=bug_report.md&title="
            )

        report_bug_button.Bind(wx.EVT_BUTTON, on_report_bug)

        donate_button = wx.Button(self, label=wx_("Donate"))

        def on_donate(_) -> None:
            webbrowser.open("https://ko-fi.com/adamws")

        donate_button.Bind(wx.EVT_BUTTON, on_donate)

        box.Add(report_bug_button, 0, wx.EXPAND | wx.ALL, 5)
        box.Add(donate_button, 0, wx.EXPAND | wx.ALL, 5)

        return box

    def get_help_section(self):
        box = wx.BoxSizer(wx.VERTICAL)
        help_message = wx.TextCtrl(
            self, style=wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_AUTO_URL | wx.HSCROLL
        )

        bold_font = wx.Font(wx.FontInfo().Bold())
        bold_attr = wx.TextAttr(wx.BLACK)
        bold_attr.SetFont(bold_font)

        regular_font = wx.Font(wx.FontInfo())
        normal_attr = wx.TextAttr(wx.BLACK)
        normal_attr.SetFont(wx.Font(wx.FontInfo()))
        normal_attr.SetFont(regular_font)

        help_message.SetDefaultStyle(bold_attr)
        help_message.AppendText(wx_("Description"))
        help_message.AppendText("\n\n")

        help_message.SetDefaultStyle(normal_attr)
        help_message.AppendText(
            self._(
                "Plugin for mechanical keyboard design. "
                "It features automatic key placement \nbased on popular "
                "layout description from "
            )
        )
        help_message.SetDefaultStyle(wx.TextAttr(wx.BLUE))
        help_message.AppendText("www.keyboard-layout-editor.com")

        def on_open_url(event) -> None:
            url = help_message.GetValue()[event.GetURLStart() : event.GetURLEnd()]
            if event.MouseEvent.LeftDown():
                webbrowser.open(url)

        help_message.Bind(wx.EVT_TEXT_URL, on_open_url)

        help_message.SetDefaultStyle(bold_attr)
        help_message.AppendText("\n\n")
        help_message.AppendText(wx_("Help"))
        help_message.AppendText("\n\n")

        help_message.SetDefaultStyle(normal_attr)
        help_message.AppendText(" \u2022 " + self._("Project website - "))

        help_message.SetDefaultStyle(wx.TextAttr(wx.BLUE))
        help_message.AppendText("https://github.com/adamws/kicad-kbplacer")

        help_message.SetDefaultStyle(normal_attr)
        help_message.AppendText("\n \u2022 " + self._("Geekhack forum thread - "))

        help_message.SetDefaultStyle(wx.TextAttr(wx.BLUE))
        help_message.AppendText("https://geekhack.org/index.php?topic=106059.0")

        dc = wx.ScreenDC()
        dc.SetFont(regular_font)

        number_of_lines = help_message.GetNumberOfLines()
        size = (0, 0)
        for i in range(0, number_of_lines):
            line = help_message.GetLineText(i)
            size_new = dc.GetTextExtent(line)
            if size_new[0] > size[0]:
                size = size_new

        margin = 20
        size = (size[0] + margin, size[1] * number_of_lines + margin)
        help_message.SetMinSize(size)

        box.Add(help_message, 0, wx.EXPAND | wx.ALL, 5)
        return box


if __name__ == "__main__":
    import sys
    import threading

    _ = wx.App(False)
    dlg = HelpDialog(None)

    if "PYTEST_CURRENT_TEST" in os.environ:
        # use stdin for gracefully closing GUI when running
        # from pytest. This is required when measuring
        # coverage and process kill would cause measurement to be lost
        def listen_for_exit():
            while True:
                input("Press any key to exit: ")
                dlg.Close(wx.ID_CANCEL)
                sys.exit()

        input_thread = threading.Thread(target=listen_for_exit)
        input_thread.daemon = True
        input_thread.start()

    dlg.ShowModal()



import logging

import pcbnew

from .board_modifier import BoardModifier, get_footprint, set_position


class TemplateCopier(BoardModifier):
    def __init__(
        self,
        board: pcbnew.BOARD,
        template_path: str,
        route_tracks: bool,
    ) -> None:
        super().__init__(board)
        self.__template = pcbnew.LoadBoard(template_path)
        self.__board_nets_by_name = board.GetNetsByName()
        self.__route_tracks = route_tracks

    # Copy positions of elements and tracks from template to board.
    # This method does not copy parts itself - parts to be positioned
    # need to be present in board prior to calling this.
    def run(self) -> None:
        footprints = self.__template.GetFootprints()

        for footprint in footprints:
            reference = footprint.GetReference()
            destination_footprint = get_footprint(self.board, reference)

            layer = footprint.GetLayerName()
            position = footprint.GetPosition()
            orientation = footprint.GetOrientation()

            if layer == "B.Cu" and destination_footprint.GetLayerName() != "B.Cu":
                destination_footprint.Flip(destination_footprint.GetCenter(), False)
            set_position(destination_footprint, position)
            destination_footprint.SetOrientation(orientation)

        if self.__route_tracks:
            tracks = self.__template.GetTracks()
            for track in tracks:
                # Clone track but remap netinfo because net codes in template
                # might be different. Use net names for remapping
                # (names in template and board under modification must match)
                clone = track.Duplicate()
                net_name = clone.GetNetname()
                net_code = clone.GetNetCode()
                net_info_in_board = self.__board_nets_by_name[net_name]
                logging.info(
                    f"Cloning track from template: {net_name}:{net_code}"
                    f"-> {net_info_in_board.GetNetname()}:"
                    f"{net_info_in_board.GetNetCode()}",
                )
                clone.SetNet(net_info_in_board)
                self.board.Add(clone)
